# ifndef ProcessComponent_h
# define ProcessComponent_h ProcessComponent_h

# include "Component.h" // parent

namespace Conjecture {
  class Page;

  /// \class ProcessComponent
  ///
  /// \brief The abstract superclass of a hierarchy of classes
  /// implementing the highest-level 'process' algorithm.
  ///
  /// Other strategy class hierarchies assume that processing consists
  /// of partitioning, identifying, and formatting. However, a
  /// creative individual may come up with an overall approach to OCR
  /// that does not follow this very general model. This Component
  /// allows for completely control over the entire OCR 'process'
  /// algorithm.
  class ProcessComponent : public Component {
  public:
    ProcessComponent();
    virtual bool execute(Page* page) = 0;
  };
}

# endif // ProcessComponent_h

