# ifndef StandardProcess_h
# define StandardProcess_h StandardProcess_h

# include "ProcessComponent.h"  // parent

namespace Conjecture {

  /// \class StandardProcess
  ///
  /// \brief Implements the most common 'process' algorithm.
  ///
  /// Almost every OCR will use the same 'process' algorithm, and
  /// will instead vary in how the cub-algorithms within 'process'
  /// are implemented.  In particular, the 'process' algorithm
  /// almost always consists of 
  ///
  ///   1) Elementitioning the image into Elements (Glyphs, Words, Lines,
  ///      Regions).  Note that only Glyph partitioning in strictly
  ///      required during this phase.
  ///
  ///   2) Identifying the unicode character associated with each
  ///      Glyph.  It is not necessary (but may be useful) for the
  ///      Glyph to already be categorized within a Word, Line, or
  ///      Region within the Page.
  ///
  ///   3) Formatting the unicode characters associated with each
  ///      Glyph. It is at this level that Glyphs MUST be organized
  ///      into Words and Lines (and possibly Regions) within the
  ///      Page, in order to ensure that spaces and newlines are
  ///      inserted in the proper places. However, as mentioned above,
  ///      identification of Words, Lines and Regions may have
  ///      occurred earlier as well.
  ///
  /// This class implements the standard 'process' algorithm discussed
  /// above.

  class StandardProcess : public ProcessComponent {
  public:
    StandardProcess();
    virtual bool execute(Page* page);
  };
}

# endif // StandardProcess_h

