using namespace std;
// C++ Dependencies
# include <iostream>
# include "GocrModule.h"  
# include "StandardProcess.h"
# include "GocrSegment.h"
# include "GocrIdentify.h"
# include "GocrFormat.h"
# include "Env.h"

// C Dependencies
# include <assert.h>
// JOB is a global variable currently needed by the gocr C code.
// It is hoped that Joerg will manage to clean up gocr so that this
// variable is not longer needed.
Job* JOB = NULL;

namespace Conjecture {

    GocrModule::GocrModule() {
	Job* job = JOB; 
	if (job == NULL ) {
	    JOB = new Job;
	    job  = JOB;
	    job_init(JOB);
	}
	this->jobIs(job);
    }

    ProcessComponent* GocrModule::createProcessComponent() const {
	return new StandardProcess();
    }
    
    SegmentComponent* GocrModule::createSegmentComponent() const {
	return new GocrSegment();
    }
    
    IdentifyComponent* GocrModule::createIdentifyComponent() const {
	return new GocrIdentify();
    }
    
    FormatComponent* GocrModule::createFormatComponent() const {
	return new GocrFormat();
    }


    void GocrModule::processArgs(int argc, const char** argv, Env* env) { 
	//cerr << "NOTE: Here processing GOCR command-line args - " << argc << endl;

        int i;
        const char *s1;
        //Job* job = this->jobRef();
	Job* job = this->job();
        assert(job);
        if (argc <= 1) {
            exit(0);
        }
#ifdef HAVE_PGM_H
        pnm_init(&argc, &argv);
#endif
        
        bool result = false;
        job->src.fname = NULL;
        /* process arguments */
        for (i = 1; i < argc; i++) {
	    //cerr << "i=" << i << " flag=" << argv[i] << " val=" << argv[i+1] << endl;
	    
            if (strcmp(argv[i], "--help") == 0) {
                result = false;
                goto EndArgs;
            }
            if (argv[i][0] == '-' && argv[i][1] != 0) {
                s1 = "";
                if (i + 1 < argc)
                    s1 = argv[i + 1];
                switch (argv[i][1]) {
                  case 'h': /* help */
                    result = false;
                    goto EndArgs;
                    break;
                  case 'i': /* input image file */
                    job->src.fname = const_cast<char*>(s1); // cast forced upon us by C code.
                    i++;
                    break;
                  case 'e': /* logging file */
                    if (s1[0] == '-' && s1[1] == '\0') {
#ifdef HAVE_UNISTD_H
                        dup2(STDOUT_FILENO, STDERR_FILENO); /* -e /dev/stdout  works */
#else
                        fprintf(stderr, "stderr redirection not possible without unistd.h\n");
#endif           
                    }
                    else if (!freopen(s1, "w", stderr)) {
                        fprintf(stderr, "stderr redirection to %s failed\n", s1);
                    }
                    i++;
                    break;
                  case 'p': /* database path */
                    job->cfg.db_path= const_cast<char*>(s1); // cast forced upon us by C code.
                    i++;
                    break;
                  case 'o': 
		    if (s1[0] == '-' && s1[1] == '\0') {
			/* default */
		    } else {
			//if (!freopen(s1, "w", stdout)) {
			//   fprintf(stderr, "stdout redirection to %s failed\n", s1);
			//}
		    };
                    i++;
                    break;
                  case 'f': /* output format */
                    if (strcmp(s1, "ISO8859_1") == 0) job->cfg.out_format=ISO8859_1; else
                        if (strcmp(s1, "TeX")       == 0) job->cfg.out_format=TeX; else
                            if (strcmp(s1, "HTML")      == 0) job->cfg.out_format=HTML; else
                                if (strcmp(s1, "XML")       == 0) job->cfg.out_format=XML; else
                                    if (strcmp(s1, "SGML")      == 0) job->cfg.out_format=SGML; else
                                        if (strcmp(s1, "UTF8")      == 0) job->cfg.out_format=UTF8; else
                                            if (strcmp(s1, "ASCII")     == 0) job->cfg.out_format=ASCII; else
                                                fprintf(stderr,"Warning: unknown format (-f %s)\n",s1);
                    i++;
                    break;
                  case 'c': /* list of chars (_ = not recognized chars) */
                    job->cfg.lc = const_cast<char*>(s1); // cast forced upon us by C code.
                    i++;
                    break;
                  case 'C': /* char filter, default: NULL (all chars) */
                    /* ToDo: UTF8 input, wchar */
                    job->cfg.cfilter = const_cast<char*>(s1); // cast forced upon us by C code.
                    i++;
                    break;
                  case 'd': /* dust size */
                    job->cfg.dust_size = atoi(s1);
                    i++;
                    break;
                  case 'l': /* grey level 0<160<=255, 0 for autodetect */
                    job->cfg.cs = atoi(s1);
                    i++;
                    break;
                  case 's': /* spacewidth/dots (0 = autodetect) */
                    job->cfg.spc = atoi(s1);
                    i++;
                    break;
                  case 'v': /* verbose mode */
                    job->cfg.verbose |= atoi(s1);
                    i++;
                    break;
                  case 'm': /* operation modes */
                    job->cfg.mode |= atoi(s1);
                    i++;
                    break;
                  case 'n': /* numbers only */
                    job->cfg.only_numbers = atoi(s1);
                    i++;
                    break;
                  case 'x': /* initialize progress output s1=fname */
                    cerr << "ERORR: -x currently disabled in C++ implementation" << endl;
                    // ini_progress(s1);
                    i++;
                    break;
                  case 'a': /* set certainty */
                    job->cfg.certainty = atoi(s1);;
                    i++;
                    break;

                  default:
                    //fprintf(stderr, "# unknown option use -h for help\n");
		    i = i;
                }
                continue;
            }
            /* argument can be filename v0.2.5 */
            else if ( (argv[i][0] != '-') || ( argv[i][1] == '\0' ) ) {
                job->src.fname = const_cast<char*>(argv[i]); // cast forced upon us by gocr code.
            }
        }
      EndArgs:
	i++;
    }

};

