# ifndef OcradModule_h
# define OcradModule_h OcradModule_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "OCRModule.h" // parent

namespace Conjecture {

    /******************************************************************
     * \class OcradModule
     *
     * \brief A wrapper for the third-party open-source Ocrad OCR
     *
     * The current implementation of Ocrad is temporary - it invokes
     * 'ocrad' in a subprocess, thus delegating all responsiblity to
     * an external program and providing no access to the internals.
     *
     * An implementation that links the Ocrad library into Conjecture,
     * similar to what was done for Gocr, will be provided when
     * someone finds time to do so.
     *
     */// *************************************************************
    class OcradModule : public OCRModule {
      public:
        // Constructors/Destructors
        OcradModule();
        
        // **************
        // Interface
        virtual ProcessComponent*     createProcessComponent() const;
        virtual SegmentComponent*     createSegmentComponent() const;
        virtual IdentifyComponent*    createIdentifyComponent() const;
        virtual FormatComponent*      createFormatComponent() const;
    };
}

# endif // OcradModule_h

