# ifndef WmhIdentify_h
# define WmhIdentify_h WmhIdentify_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "IdentifyComponent.h" // parent
# include <vector>
# include <map>

namespace Conjecture {
    class Glyph;
    class PixFilt;
    class GlyphFeatures;

    // ##################################################################
    /// \class WmhIdentify
    ///
    /// \brief
    ///
    /// ...
    // ##################################################################
    class WmhIdentify : public IdentifyComponent {
      public:
        // Types
        typedef std::map<wchar_t,GlyphFeatures*> Charset;

        // Constructors/Destructors
        WmhIdentify();
        virtual ~WmhIdentify();
        
        // **************
        // Accessors
        inline const Charset & charset() const { return this->_charset; }
        inline const PixFilt*  stdfilt() const { return this->_stdfilt; }
        inline const std::vector<PixFilt*> & filters() const { return this->_filters; }
        
        // **************
        // Input/Output
        
        // **************
        // Interface
        virtual bool execute(Page* page);

        virtual bool execute(Glyph* glyph);
        
      protected:
        // **************
        // Accessors
        inline void            charsetIs(const Charset & charset) { this->_charset = charset; }
        inline void            stdfiltIs(PixFilt* stdfilt) { this->_stdfilt = stdfilt; }
        inline void            filtersIs(const std::vector<PixFilt*> & filters) { this->_filters = filters; }
        
        // **************
        // Methods 
        
      private: 
        // **************
        // Accessors
        inline Charset &       charsetRef() { return this->_charset; }
        inline PixFilt* &      stdfiltRef() { return this->_stdfilt; }
        inline std::vector<PixFilt*> & filtersRef() { return this->_filters; }
        
        // **************
        // Methods 
        
        // **************
        // State

        /// The collection of all recognized unicode characters, and
        /// associated GlyphFeatures description.
        ///
        /// This field is critical to the proper functioning of
        /// WmhAlgorithms. Every Glyph is analyzed and a set of
        /// GlyphFeatures is established. Furthermore, every possible
        /// output character is assigned a "perfect" GlyphFeatures
        /// instance. The GlyphFeatures class provides a comparison
        /// method that determines the degree of similarity between
        /// desired and given. From this, we can establish the
        /// character that is most likely to be correct.
        Charset         _charset;

        /// The 101-->111 filt 
        ///
        /// This is kept separate from 'filters' because it is often
        /// applied before anything else occurs.
        PixFilt*        _stdfilt;

        /// We maintain a collection of filters to be applied to each glyph
        std::vector<PixFilt*> _filters;
    };
}

# endif // WmhIdentify_h

