# ifndef WmhSegment_h
# define WmhSegment_h WmhSegment_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "SegmentComponent.h" // parent

namespace Conjecture {
    
    // ##################################################################
    /// \class WmhSegment
    ///
    /// \brief Responsible for sub-dividing the Page into Glyphs (and,
    /// optionally, Words, Lines and Parts).
    ///
    /// The abstract superclass of a hierarchy of classes implementing
    /// the 'Segment' algorithm, which is minimally responsible for
    /// identifying the sub-regions of the Page that represent
    /// individual Glyphs. The algorithm may do more than just
    /// identify Glyphs however (i.e. may also identify Words, Lines
    /// and/or Regions too).
    // ##################################################################
    class WmhSegment : public SegmentComponent {
      public:
        WmhSegment();
        virtual bool execute(Page* page);
    };
}

# endif // WmhSegment_h

