# ifndef SomeClass_Bitmap_h
# define SomeClass_Bitmap_h SomeClass_Bitmap_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################
// Class Champion: Bruno

# include "Root.h"  // parent
# include "Coord.h"
# include "ImageFilter.h"

namespace Conjecture {

    typedef unsigned char Pixel; // TODO: should this be a class?

    // ##################################################################
    /// \class ImageB
    ///
    /// \brief Image class.
    ///
    /// ...
    // ##################################################################
    class Bitmap : public Root {

    public:
        // Constructors/Destructors
        Bitmap(uint32_t width, uint32_t height);

        Bitmap(const std::string &filename);

        ~Bitmap();

        // **************
        // Accessors
        inline const Pixel operator() (uint32_t x, uint32_t y) const {
            return ( (_pixels[at(x,y)] & (1 << (x & 0x07))) == 0 ? 0 : 1);
        }

        inline const Pixel mask(uint32_t x, uint32_t y) const {
            return ( (_mask[at(x,y)] & (1 << (x & 0x07))) == 0 ? 0 : 1);
        }

        inline const uint32_t width() const {
            return this->_width;
        }

        inline const uint32_t height() const {
            return this->_height;
        }

        // **************
        // Input/Output

        // **************
        // Interface

        /**
         *
         */
        Bitmap clone(void) const;

        /**
         *
         * @
         * @param a Top left pixel of area to clone.
         * @param b Bottom right pixel of area to clone.
         * @return A new bitmap
         */
        const Bitmap clone(Coord a, Coord b) const;

        /**
         * Applies a filter to this image.
         *
         * @param imfilter The image filter to apply to this image.
         */
        bool filter(ImageFilter &imfilter);

        /// Unit testing method.
        ///
        /// This static method should create instances of the class (and
        /// instances of any other class necessary) and perform tests to
        /// ensure that all methods within the class are working as
        /// expected.
        static void test(int argc = 0, const char* argv[] = NULL);

    protected:
        // **************
        // Accessors

        // **************
        // Methods

    private:
        // **************
        // Accessors

        // **************
        // Methods

        /**
         * Returns the array position of a given pixel.
         *
         * @param x X coordinate
         * @param y Y coordinate
         */
        inline uint32_t at(uint32_t x, uint32_t y) const {
            return (y * _linesize + (x >> 3));
        }

        // **************
        // State

        typedef uint32_t PixelBlock; // TODO: this should be platform dependent

        uint32_t        _width;
        uint32_t        _height;
        uint32_t        _linesize;
        PixelBlock      *_pixels;
        PixelBlock      *_mask;

    };
}

# endif // SomeClass_h

