# include <iostream>
# include <iomanip>
# include "Glyph.h"  
# include "Image.h"
# include "Line.h"

# include "pgm2asc.h"
# include "gocr.h"    // Box

// ******************************************************************
// CONSTRUCTORS

# include <iostream>
using namespace std;

namespace Conjecture {

    Glyph::Glyph(Element* parent, void* data) : Element(parent, Coord(0,0), Coord(0,0)) {
	// NOTE: The Coord(0,0) in the init list is rather dirty, but we
	// do promptly initialize 'topleft' and 'size' within this
	// constructor as long as 'data' is provided!
	Box* box;
	if ( data ) {
	    // Horrible type-violation - temporary.
	    box = (Box*)data;
	    this->topleftRef().x = box->x0;
	    this->topleftRef().y = box->y0;
	    this->bottomrightRef().x = box->x1;
	    this->bottomrightRef().y = box->y1;
	} else {
	    //cerr << "ERROR: Creation of Glyph without underlying Box is currently not safe!" << endl;
	    box = (Box*) malloc_box(NULL);
	}
	this->boxIs(box);
	
	// C++ state
	this->bestIs(-1); // index of best char so far
    }
    
    // Code copied from 'malloc_box'
    Glyph::Glyph(Glyph* glyph) : Element(glyph->parentRef(), Coord(0,0), Coord(0,0)) {
	Box* box = (Box*) malloc_box(glyph->boxRef());
	this->boxIs(box);
    }

    void Glyph::writeText(ostream& os) const {
	// This naive implementation doesn't usually to the proper
	// thing - subclasses reimplement as necessary.
	os << (char)this->unicode();
    }

    GlyphFeatures::LineType Glyph::lineType() const {
	GlyphFeatures::LineType result = GlyphFeatures::LTUnknown;
	const Line* line = this->parent()->asLine();
	int debug = 0;

	if ( line ) {
	    // The Glyph belongs to a line, so we have access to
	    // baseline, midline, ascent and descent.
	    //
	    // IMPORTANT: I am fudging the baseline test slightly.
	    // The test image I am using tends to sometimes
	    // have glyphs with bottom y values one pixel lower than
	    // the line baseline and they are often misidentified as
	    // Descender when they are Base.  Hence the +1 below.
	    if ( this->topY() < line->midY() ) {
		// Ascender or Total
		result = ( this->bottomY() > line->baseY() ) ? GlyphFeatures::Total : GlyphFeatures::Ascender;
		DBG(1,"Found Ascender");
	    } else if ( this->bottomY() > line->baseY()+1 ) {
		// Descecender (see notes above for +1 fudge discussion)
		result = GlyphFeatures::Descender;
		DBG(1,"Found Descender");
	    } else {
		// Base
		result = GlyphFeatures::Base;
		DBG(1,"Found Base");
	    }

	} else {
	    cerr << "ERROR: No line found" << endl;
	    cerr << "Parent " << this->parent() << endl;
	}
	return(result);
    }

    void Glyph::addChar(const Glyph::Char& chr) {
	std::vector<Char>& set = this->charset();
	
	// Add the char
	set.push_back(chr);
	
	// Determine if the new chr is better than the previous best
	int best = this->best();
	if ( best == -1 ) {
	    // wasn't set before, so size must be 1 and best is index 0.
	    best = 0;
	} else if ( set[best].weight < chr.weight ) {
	    best = set.size()-1;
	}
    }

    void Glyph::printSummary(ostream& os, const string& indent, int index) const {
	os << indent << "Glyph ";
	if ( index >= 0 ) { os << right << setw(4) << index; }
	const Box* box = this->box();
	
	os << ": "
	   << " id="   << left << setw(10) << this->id().c_str()
	   << " tl="   << this->topleft().str(3)
	   << " br="   << this->bottomright().str(3)
	   << " sz="   << this->size().str(2)
	   << " m1="   << setw(3) << box->m1
	   << " m2="   << setw(3) << box->m2
	   << " m3="   << setw(3) << box->m3
	   << " m4="   << setw(3) << box->m4
	   << " c="    << box->c
	   << endl;
    }
    
    void Glyph::writeImage(const std::string& filebase, const ImageArgs& config) {
	// Obtain the page image.
	const Image* image = this->pageImage();
	// Make a copy of the glyph's subregion, appropriate filterd, magnified and thresholded.
	Image gim(image, this->topleft(), this->bottomright(), config);
	// Write the image to disk.
	bool verbose = false; // here preparing for future support
	gim.writeGrey(filebase,verbose);
    }
    
    void Glyph::test(int argc, const char* argv[]) {
	cerr << "ERROR: Not unit test for Glyph - FIX THIS!" << endl;
    }

};
