# include <iostream>
# include <sstream>
# include "Line.h"  
# include "Env.h" // test()
# include "ImageArgs.h"
# include "Glyph.h"
# include "Image.h"

using namespace std;

namespace Conjecture {

    Line::Line(Element* parent, const Coord& topleft, const Coord& bottomright, u2 bottom, u2 base, u2 mid, u2 top) : Element(parent, topleft, bottomright) {
	this->topYIs(top);
	this->midYIs(mid);
	this->baseYIs(base);
	this->bottomYIs(bottom);
    }

    void Line::printSummary(ostream& os, const string& indent, int index) const {
	os << "Line: bottom=" << this->bottomY() 
	   << " base=" << this->baseY()
	   << " mid=" << this->midY()
	   << " top=" << this->topY()
	   << endl;
	if ( index >= -1 ) {
	    this->Element::printSummary(os,indent,index);
	}
    }
    
    void Line::writeText(ostream& os) const {
	
	int i = 0;
	for (ElementList::const_iterator it = this->parts().begin(); it != this->parts().end(); ++it) {
	    if ( i > 0 ) { os << " "; }
	    (*it)->writeText(os);
	    ++i;
	}
	// Lines always end within newlines
	os << endl;
    }

    void Line::test(int argc, const char* argv[]) {
	
	// Create an Env that loads up test.pnm and writes output to test.ocr
	static const int locargc = 5;
	static const char* locargv[locargc] = { "exec", "-i", "test.pnm", "-o", "test.ocr" };
	Env * env = new Env(locargc,locargv);
	
	// Segment the current page 
	string fname = env->infile();
	cerr << "NOTE: Loading " << fname << endl;
	Page* page = env->currentPage();
	page->segment();
	
	// Assuming that segmentation has placed Glyphs into
	// Lines, we write the images for the first line.
	Line* line = page->firstElement()->asLine();
	if ( line ) {
	    //const Image* image = line->pageImage();
	    
	    // Loop over all imags in the line
	    GlyphLoopStart(line, glyph) {
		// Do something here!???
		glyph->printSummary();
	    } GlyphLoopEnd;
	} else {
	    cerr << "ERORR: Did not find a line" << endl;
	}
    }
};
