# ifndef Word_h
# define Word_h Word_h

// Dependencies
#include "Element.h" // parent

// Declaration
namespace Conjecture {

    /** ############################################################
     * \class Word
     *
     * \brief A sub-region of a Line consisting entirely of Glyphs.
     *
     * A Word is separated from other Words within the Line by more
     * horizontal space than between Glyphs within a Word.
     *
     * The order in which Words are partitioned (before or after
     * Glyphs are partitioned) depends on the algorithm, and use
     * client-controlled. Note that accurate Word partitioning can
     * improve Glyph identification in situations where the
     * identification algorithm incorporates natural-language
     * knowledge (spell checkers, etc.). However, it is expected that
     * the Word class will be relatively unimportant except during
     * output (to insert spaces), and thus not even strictly necssary
     * (space insertion can be performed without requiring Word
     * instances to be created).
     *
     */// ##########################################################

    class Word : public Element {
      public:
        // Constructors/Destructors
        Word(Element* parent, const Coord& topleft, const Coord& bottomright);
        
        /**
         * Convert an Element to Word if it is a Word
         */
        virtual Word* asWord()             { return this; }
        virtual const Word* asWord() const { return this; }
        
        virtual void printSummary(std::ostream& os = std::cerr, const std::string& indent = "", int index = -1) const;

	/**
	 * Writes a textual representation of this element to given output stream
	 *
	 */
	virtual void writeText(std::ostream& os) const;
        
        /**
         * Unit testing method.
         *
         * This static method should create instances of the class (and
         * instances of any other class necessary) and perform tests to
         * ensure that all methods within the class are working as
         * expected.
         *
         */
        static void test(int argc = 0, const char* argv[] = NULL);
        
        // Accessors
        virtual int type() const { return Type; }
        
      protected:
        // Accessors
        
        // Methods
        
      private:
        // Accessors
        
        // Methods
        
        // State
        static const ElementType Type = ELEMENT_WORD;
    };
};

# endif // Word_h

