using namespace std;
# include <iostream>
# include "BitInfo.h"

// Created 2006/06/10 by wade

// Include

namespace Conjecture {

    BitInfo::BitInfo(BitOverlay::Layout layout) : BitOverlay(layout) {

	// Number of regions
	BitCount regcnt  = 0;

	// Number of pixels in curent region
	int      regcurr = 0;

	// Number of pixels in largest region
	int      regmax     = 0;

	// True if currently processing a "on" region
	bool     active  = false;

	// The first high bit that is "on"
	BitPos high = BitOverlay::npos;

	// The first low  bit that is "on"
	BitPos low  = BitOverlay::npos;

	// The set of all regions (sizes).  There can be at most 16.
	unsigned short* regions = this->regionsRef();

	// Current bit position being analyzed
	for (int i = 31; i >= 0; --i) {
	    // Establish the pixel to look at
	    unsigned mask    = (1<<i);

	    if ( layout&mask ) {
		// ******************
		// Current bit is "on"
		
		// We increment the pixel count for the current region
		++regcurr;

		// Indicate that we are processing an "on" region
		//   FUTURE FIX: Can we avoid having to set this every time?
		active = true;

		// If we haven't already existablished high, this is it.
		if ( high == BitOverlay::npos ) {
		    high = i;
		}

		// This is the lowest "on" bit so far
		low = i;

	    } else if ( active ) {
		// *******************
		// The current pixel is the first "off" bit after
		// a region of "on" bits.

		// No longer in an "on" region
		active = false;

		// CODE TANGLE #1:
		// Check if the current region is the largest
		if ( regcurr > regmax ) { regmax = regcurr; }
		// Remember the size of this region
		regions[regcnt] = ((i+1)<<8) | regcurr;
		// Advance number of regions
		++regcnt;
		// Reset the region count
		regcurr = 0;
	    }
	}
	// If regcurr is non zero, it means we have a final region
	// to process:
	if ( regcurr > 0 ) { 
	    // CODE TANGLE #1
	    if ( regcurr > regmax ) { regmax = regcurr; }
	    regions[regcnt] = regcurr;  // start of region is (0<<8) = 0 so noop
	    // do not need to increment regcnt or reset regcurr - we're done
	}

	// ******************
	// Initialize the state
	this->largestRegionIs(regmax);
	this->regionCountIs(regcnt);
	this->firstLowIs(low);
	this->firstHighIs(high);
    }
    
    // Interface

    ostream& BitInfo::print(ostream& os, const string& indent) const {
	os << indent << Str(this->bits()) << " (" << hex << (int)this->bits() << dec << ")" << endl;
	os << indent << "  High    : " << this->firstHigh() << endl;
	os << indent << "  Low     : " << this->firstLow() << endl;
	os << indent << "  Regions : (" << this->regionCount() << ")";
	for (unsigned i = 0; i < this->regionCount(); ++i) {
	    os << " ";
	    unsigned short regcnt = this->regions()[i];
	    int      start = regcnt >> 8;
	    unsigned size  = regcnt & 0xff;
	    int      end   = start+size-1;
	    if ( size == this->largestRegion() ) {
		os << "*";
	    }
	    os << size << "=" << end;
	    if ( start != end ) {
		os << "-" << start;
	    }
	}
	os << endl;
	return os;
    }


    // The Unit Testing Method!!!
    void BitInfo::Test(int argc, const char* argv[]) {
	try {
	    BitInfo b1(3);
	    b1.print();
	    
	    BitInfo b2(1392);
	    b2.print();
	    
	    BitInfo b3(7593920);
	    b3.print();
	} catch (const string& e) {
	    cerr << "Here with " << e << endl;
	}
    }


};

