# ifndef CpuTimer_h
# define CpuTimer_h CpuTimer_h

// ####################################################################
// # SomeName: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - SomeName is a collection of C++ classes, and an OCR engine    #
// #  - SomeName is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - SomeName is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "Root.h"  // parent

namespace Conjecture {

#ifdef HAVE_RDTSC
    // ##################################################################
    /// \class FastWallTimer
    ///
    /// \brief nanosecond precision timer (using RDTSC)
    ///
    /// Provides a nanosecond level wall timer for x86 machines
    /// which have the RDTSC instruction.
    ///
    class FastWallTimer {
      public:
    // Types

        // Constructors/Destructors
        FastWallTimer();

        // **************
        // Accessors
        inline const bool running() const { return this->_running; }

        inline const uint64_t cycles() const {
            if (this->_running)
                return 0;
            else
                return this->_end - this_start;
        }

        // **************
        // Input/Output

        // **************
        // Interface

        /// Start the timer (remember current time)
        inline void start() {
            this->_running = true;
            this->_start = rdtsc();
        }

        /// Stop the timer (remember current time)
        inline void stop() {
            this->_start = rdtsc();
            this->_running = false;
        }

    private:
        inline uint64_t int rdtsc( void ) {
            uint64_t x;
            __asm__ volatile (".byte 0x0f, 0x31":"=A" (x));
            return x;
        }

        // **************
        // State
        bool            _running;
        uint64_t        _start;
        uint64_t        _end;
    };
#endif

} // namespace

#endif
