<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
  <title>Conjecture - Modules</title>
  <meta http-equiv="content-type" content="text/html; charset=ISO-8859-15">
  <!-- link rel="icon" href="conjecture.ico" type="image/x-icon" -->
  <!-- link rel="shortcut icon" href="conjecture.ico" type="image/x-icon" -->
  <link rel="stylesheet" href="style.css" type="text/css" title="plain">
</head>
<body>
<table class="nopad" align="left" border="0" cellpadding="0"
 cellspacing="0" width="100%">
  <tbody>
    <tr>
      <td class="nopad">
<?php
 $mainvalue='Documentation';
 $subvalue='Modules';
 include_once("header.php");
?>
      </td>
    </tr>
    <tr>
      <td>

      <table class="nopad" align="left" border="0" cellpadding="0" cellspacing="0" width="100%">
        <tbody>
          <tr valign="top">
            <td class="content">

<div id="overview">
  <h1>An Overview of Modules</h1>
  <blockquote>
    <p>You may want to read up on <a href="components.php">Components</a>
    before reading about Modules. This documentation discusses how
    Conjecture combines a collection of Components into a self-contained
    OCR; a Module.

    <p>Conjecture provides support for comparing, sharing and interaction
    between an unlimited number of OCRs. This is accomplished by defining
    a OCR Module hierarchy. Each class in this hierarchy represents a
    fully-functional OCR capable of taking an image as input and producing
    text as output. Each such class must deal with input image processing
    (noise removal, orientation adjustment, etc.), segmentation into
    glyphs, identification of unicodes for each glyph, the production of
    properly formatted output, and everything else that goes into an OCR.
    However, these classes almost never do so directly. Instead, they
    almost always delegate responsibility for implementing individual
    issues to a specific subclass within the Component hierarchy
    associated with that issue.

    <p>In fact, classes in the OCR Module hierarchy are usually very simple,
    and are almost always automatically generated by the Conjecture
    infrastructure based on an entry in a special input file (see the
    discussion on <code>Conjecture.modules</code>).

    <p>The OCR Module hierarchy has an abstract class called, not
    surprisingly, OCRModule, as its root. This class defines a special
    interface consisting of a collection of Factory Methods (methods
    responsible for creating objects). One Factory Method exits for each
    Component that Conjecture has identified and formalized. The OCRModule
    class also defines one field for each component, and during creation
    uses the factory methods in order to assign instances to each of these
    fields.

    <p>Subclasses of the abstract OCRModule class represent concrete,
    fully-functional OCR implementations. They usually consist solely of
    definitions of the factory methods required by the OCRModule
    superclass. If the OCR in question does not need a particular
    Component, the associated factory method simply returns NULL. Each
    factory method specifies which Component subclass to create, and thus
    establishes which implementation of each Component is used by this
    particular OCR Module.

    <p>The Conjecture framework (specifically, the Env class) maintains a
    special 'ocr' field that is initialized to an instance of a subclass
    of OCR based on the <code>-A</code> command-line flag. This field
    effectively acts as a container for a collection of Composite subclass
    instances. Anywhere in the framework, if one wants to perform
    segmentation, they can ask for the 'ocr' field of Env, and from this
    field invoke the public 'segment()' method, which returns a subclass
    of SegmentComponent. The 'execute(Page*page)' method can then be
    invoked on this object in order to perform segmentation. All without
    having to know anything whatsoever about the actual implementation
    details. This design allows us to complete separate individual
    components from one another, making for a flexible plug-and-play
    paradigm.

    <p>The current class diagram for the OCRModule hierarchy is:

    <div class="fragment"><pre class="fragment">
                        OCRModule
                           |          
      .--------------------^-----.-----------. --- ...
      |                |         |           |             |
 DefaultModule   GocrModule  OcradModule  WmhModule         ...
    </pre></div>

    <ul>

      <li> The <code>DefaultModule</code> class describes the default
      Conjecture implementation, and will represent the current "best"
      overall strategy.

      <li> The <code>GocrModule</code> class relies on the GOCR library,
      and provides facilities for having Conjecture data-structures like
      Page, Glyph, etc. interact with underlying GOCR data-structures like
      Job and box.

      <li> The <code>OcradModule</code> provides access to the Ocrad OCR,
      but currently via a sub-process execution (as opposed to linking the
      Ocrad source code into Conjecture). This will be changed as
      Conjecture evolves.

    </ul>

    <div id="modules_file">
      <h2> The <code>config/Conjecture.modules</code> File </h2>
      <blockquote>

    	<b>NOTE: The name of this file will be changing soon!</b>

    	<p> The <code>config/Conjecture.modules</code> file provides a
    	description of every OCR Module currently supported by Conjecture. To
    	add a new OCR Module, a user simply edits this file and adds another
    	record in the same format as those already present. This file is
    	expected to grow as more and more Modules are implemented, so we do
    	not present a comphrensive summary here. Instead, we'll provide a
    	small sample file and discuss the syntax for it:

    	<pre>
  MODULE default CLASS DefaultModule
            COMPONENT Process  CLASS StandardProcess;
     extern COMPONENT Segment  CLASS GocrSegment;
     extern COMPONENT Identify CLASS GocrIdentify;
     extern COMPONENT Format   CLASS GocrFormat;

  MODULE gocr CLASS GocrModule
     extern COMPONENT Process  CLASS StandardProcess;
            COMPONENT Segment  CLASS GocrSegment;
            COMPONENT Identify CLASS GocrIdentify;
            COMPONENT Format   CLASS GocrFormat;

  MODULE ocrad CLASS OcradModule
            COMPONENT Process  CLASS OcradProcess;
     no     COMPONENT Segment;
     no     COMPONENT Identify;
     no     COMPONENT Format;

  MODULE wmh CLASS WmhModule
     extern COMPONENT Process  CLASS StandardProcess;
            COMPONENT Segment  CLASS WmhSegment;
            COMPONENT Identify CLASS WmhIdentify;
            COMPONENT Format   CLASS WmhFormat;
    	</pre>

    	<p> There are four Modules defined in this sample file, with the
    	names 'default', 'gocr', 'ocrad' and 'wmh'. Note that the public
    	name appears immediately after 'MODULE' keyword, and is not
    	necessarily the same as the class name (which appears after the
    	'CLASS' keyword). When invoking <code>conjecture</code> at the
    	command line, the <code>-A<code> flag accepts a module name as a
    	value.

    	<p> Each OCR Module is fully described by specifying which class
    	it will use to implement each Component. For example, GocrModule
    	uses the StandardProcess subclass of ProcessComponent to implement
    	the 'process' issue, and uses Gocr-specific sublcasses of
    	SegmentComponent, IdentifyComponent and FormatComponent to deal
    	with segmentation, identification and formatting respectively.

      </blockquote>
    </div>

    <div id="ocrgen">
      <h2> The <code>bin/ocrgen</code> Script </H3>
      <blockquote>

    	<p>The <code>ocrgen</code> script reads the
    	<code>Conjecture.modules</code> file and verifies the existence of an
    	appropriate subdirectory of <code>src/custom</code>, and source/header
    	files for the Module and Component classes provided by the Module.

    	<p> ... more here ...
      </blockquote>
    </div>

  </blockquote>
</div>

<div id="creating">
  <h1>Creating a new OCR Module</h1>
  <blockquote>
  </blockquote>
</div>

<div id="existing">
  <h1>OCR Modules Currently Available in Conjecture</h1>
  <blockquote>
  </blockquote>
</div>

            </td>

            <!-- begin side bar -->
            <td class="right" width="200">
<?php
 include_once("sidebar.php");
?>

            </td>

            <!-- end side bar -->

          </tr>
      </table>
      </td>
    </tr>
    <tr>
      <td class="nopad">
<?php
 include_once("footer.php");
?>
      </td>
    </tr>
  </tbody>
</table>


</body>
</html>
