package Wmh;
use strict;

=head1 NAME

Wmh -

=head1 SYNOPSIS

use Wmh ();

=head1 DESCRIPTION


=head1 NOTES


=head1 EXAMPLES


=head1 IMPLEMENTATION

=cut

use Carp;                      # everyone gets Carp!
use IO::File;                  # everyone gets IO::File!
use IO::File;
use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK);
@ISA = qw();
$VERSION = '0.01';

####################### Package Global Variables ###################
use vars qw();


use vars qw(@ISA  %PrintFormat $ElementHash $ElementCount $RightMargin);
@ISA = qw(Exporter);

use Carp;
use IO::File;

use Wmh::Extensions;
# require 'Wmh/Extensions.pl';

# See end of file for pod documentation of module purpose

####################### Package Global Variables ###################

# Default formatting for print_string
%PrintFormat = (  'keyform'   => '%-20s',
                  'idxform'   => '[%3d]',
                  'keydent'   => '    ',
                  'idxdent'   => '    ',
                  'classdent' => '    ',
                  'indent'    => '    ',
                  'maxdepth'  => 8,
           );

# This hash maintains a list of all values printed so far.  This allows
#   the printing to avoid reprinting multiple copies of the identical
#   element.
$ElementHash = {};

# This integer indicates how many elements have been printed so far.
#    Currently, it is flat (i.e. not dependent on how deeply nested the
#    printing goes), but eventually I might make the element numbering
#    scheme more sophisticated
$ElementCount = 0;

# This integer dictates the right margin (where annotations should occur)
$RightMargin = 80;

############################################################################

######
# I can make copies of atomics (perl does it automatically :-),
#  SCALAR refs, ARRAY refs and HASH refs
#
# Cannot yet, but certainly will eventually, handle self-referencing structurs.
#
# Use a depth of -1 for arbitrarily deep copy (although this
# has the potential to cause an infinite loop - be careful!)
sub copy {
    my ($var, $depth) = @_;
    $depth = 1 unless defined($depth);
    my $type = ref($var);
    my $result;

    my $class;

    if ( $var =~ m/^([^=]+)=([A-Z]+)\(/ ) {
        # Class definition
        $class = $1;
        $type = $2;
    }

    if ( $type eq '' ) {
        $result = $var;

    } elsif ( $type eq 'SCALAR' ) {
        $result = \$var;  # since $var is a copy of the var passed in :-)

    } elsif ( $type eq 'ARRAY' ) {
        # Keep this == 0 (i.e. dont make it <= 0, or -1 won't work for infinite depth)
        if ( $depth == 0 ) {
            $result = [@$var];  # new reference pointing to same things as $var
        } else {
            my @arr;
            my $d = $depth - 1;
            foreach (@$var) {
                push(@arr, &copy( $_, $d ));
            }
            $result = \@arr;
        }
    } elsif ( $type eq 'HASH' ) {
        # Keep this == 0 (i.e. dont make it <= 0, or -1 won't work for infinite depth)
        if ( $depth == 0 ) {
            $result = {%$var};  # new reference pointing to same things as $var
        } else {
            my %hash;
            my $key;
            my $d = $depth - 1;
            foreach $key (keys(%$var)) {
                if ( ref($_) ne '' ) {
                    printf(STDERR "ERROR: &copy cannot handle non-atomic keys (how deeply should I copy the key structure?\n");
                    return undef;
                }
                $hash{$key} = &copy( $var->{$key}, $d );
            }
            $result = \%hash;
        }
    } elsif ( $type eq 'REF' ) {
        printf(STDERR "&copy(REF) not yet implemented\n");

    } elsif ( $type eq 'GLOB' ) {
        printf(STDERR "&copy(GLOB) not yet implemented\n");

    }

    if (defined($class)) {
        bless $result, $class;
    }

    return $result;
}


# Return an array of values obtain from a hash reference and array of keys,
# providing default values (also provided) if hash values are not defined.
#
#  Example:
#     &hashValues( $hash, ['key1', 'key2', 'key3'], [{}] );
#     &hashValues( $hash, ['key1', 'key2', 'key3'], [{}, [], 2] );
sub hashValues {
    my ($hash, $keys, $defaults ) = @_;

    if ( $#$defaults == 0 ) {
        # Only one default - assume that shallow copies of it are to be used for all
        # defaults
        for (1..$#$keys) {
            push(@{$defaults}, &copy($defaults->[0]));
        }
    }

    my @result;
    my $i;
    my $key;
    my $val;
    for $i (0..$#$keys) {
        $key = $keys->[$i];
        $val = defined($hash->{$key}) ? $hash->{$key} : $defaults->[$i];
        push(@result, $val);
    }

    return @result;
}

sub permute {
    # Return a permutation of the given array
    my (@arr) = @_;

    my @res;
    while (@arr > 0) {
        my $i = int(rand() * @arr);
        push(@res, $arr[$i]);
        splice(@arr, $i, 1);
    }
    return (wantarray) ? @res : \@res;
}

#########:
# Purpose: Return a string representing my internal structure
# Args   : $self : ANYTHING, %format : list
# Notes  :
# Examples:
#
#      @a = (1,2,3,4);
#      Wmh::print_string(\@a);
#
# Return : text

sub print_string
{
    my $var = shift;
    my %format = @_;

    $ElementHash = {};
    $ElementCount = 0;

    %format = &_generate_print_format(%format);
    return( &_print_string($var,'',\%format,$format{'maxdepth'}) );
}


# Given a list of strings, return a (multi-line) string representing
# the list placed into as many columns as will fit in $width
sub columnize {
    my ($list, $width, $colwidths) = @_;
    $width = 80 unless defined($width);
    $colwidths = [] unless defined($colwidths);

    # Obtain the length of each string and, in particular,
    # the length of the longest string.
    my @widths;
    my $max = 0;
    my $len;
    foreach (@$list) {
        $len = length($_);
        $max = $len if ($len > $max);
        push(@widths, $len);
    }
    ++$max;  # add one for space

    # Determine the number of columns into which
    # the list can be placed
    my $cols = $width / $max;
    $cols = 1 if ($cols < 1);

    # Create a string of $col columns
    my $result = '';
    my ($i,$j,$k);
    my $upper = $#$list/$cols;
    for $i (0..$upper) {
        for $j (0..$cols-1) {
            $k = $i*$cols+$j;
            $result .= sprintf("%-${max}s", $list->[$k]);

        }
        $result .= "\n";
    }

    ++$k;
    if ( $k <= $#$list ) {
        for $i ($k..$#$list) {
            $result .= sprintf("%-${max}s", $list->[$i]);
        }
        $result .= "\n";
    }

    return $result;
}

sub listWidth {
    # Returns the width of the widest string in given ARRAY $list
    my ($list) = @_;
    my $width = 0;
    my $l;
    foreach (@$list) {
        $l = (defined($_)) ? length($_) : 0;
#       print "Length $l for $_\n";
        $width = $l if ( $l > $width );
    }
    return $width;
}

sub listsAsColumns {
    # Given a list of array references, place them
    my ($delim, @lists) = @_;

    my @widths;
    my @sizes;
    my $maxsize = 0;
    my $size;

    foreach (@lists) {
        push(@widths, &listWidth($_));
#       printf "Width is %d\n", $widths[$#widths];
        $size = $#$_ + 1;
        push(@sizes, $size);
        $maxsize = $size if ( $size > $maxsize );
    }

    my ($i,$j);
    my @result;
    my $line;
    my $width;
    my $str;
    foreach $i (0..$maxsize-1) {
        $line = '';
        foreach $j (0..$#lists) {
            $width = $widths[$j];
            $str = $lists[$j]->[$i];
            $line .= sprintf("%s%-${width}s", (($j==0)?'':$delim), (defined($str) ? $str : ''));
        }
        push(@result, $line);
    }
    return (wantarray) ? @result : join("\n", @result);
}

sub fixedLengthStr {
    my ($str, $len, $terminator, $orient, $pad) = @_;
    $terminator = "\$" unless defined($terminator);
    $len = 10 unless defined($len);
    $orient ||= 'left';
    $pad ||= ' ';

    # This function returns a string of a guaranteed length $len.
    # If $str has more characters than $len + length($terminator), then
    # it is truncated so that the initial portion of $str, when
    # prepended to $terminator, will have length exactly $len.
    #
    # If $str has fewer-equal characters than $len, then it is padded
    # to the left and/or right depending on the $orient argument, which
    # should be either 'left', 'right' or 'center'.

    my $termlen = length($terminator);
    $len = $termlen if ( $len < $termlen );

    my $slen = length($str);
    if ( $slen > $len ) {
        my $l = $len - $termlen;
        $str =~ s/^(.{$l}).*/$1/;
        $str .= $terminator;

    } elsif ( $slen == $len ) {
        # Nothing needs to be done.
    } else {
        # Str must be padded (on left, right or both depending on $orient)
        my $wiggle = $len - $slen;
        if ( $orient eq 'center' ) {
            my $left = int($wiggle/2);
            my $right = $wiggle - $left;
            $str = ($pad x $left) . $str . (' ' x $right);
        } elsif ( $orient eq 'right' ) {
            $str = ($pad x $wiggle) . $str;
        } elsif ( $orient eq 'left' ) {
            $str .= ($pad x $wiggle);
        } else {
            die("ERROR: Invalid orientation request '$orient' in Wmh::fixedLengthStr");
        }
    }
    return $str;
}


# Returns a sprintf format that will generate every element in the given list, if
# possible (some lists cannot be generated by a single format, in which case '%s'
# is returned)
#
# For example:
#    qw(01 02 10 99)  returns %02d
#    qw(01 002 3)     returns %s
#    qw(1 12 392)     returns %d
sub sprintfFormatFor(\@) {
    my @lengths = ();
    my $len;
    my $min = 10000;
    my $max = 0;
    my $leadzero = 0;

    # iterator over each element determining length and whether it has a
    # leading zero
    foreach (@_) {
        $len = length($_);
        $min = $len if ($len < $min);
        $max = $len if ($len > $max);
        ++$leadzero if ( m/^0./ );
        push(@lengths, $len);
    }

    my $result;
    if ( $min == $max ) {
        $result = "%0${min}d";
    } elsif ( $leadzero > 0 ) {
        $result = '%s';
    } else {
        $result = '%d';
    }

    return $result;
}

sub System {
    # %options are:
    #    error   --> error string to use
    #    indent  --> indent to use when printing errors
    #    notedent --> additional spacing to put between 'NOTE' and the text.
    #    exec    --> execute (default is 1, but can be set to 0 to disable)
    #    die     --> die on errors
    #    quiet   --> boolean true means don't print out command
    #    message --> string to print for diagnostics
    #    verify  --> boolean true means prompt for execution before executing
    my ($command, %options) = @_;

    chomp($command);
    my $error = $options{error} || "command failed";
    my $indent = $options{indent} || '';
    my $notedent = $options{notedent} || '';
    my $message = $options{message} || "Executing '$command'";
    my $verbose = $options{verbose} || 0;

    # If exec is defined but is 0, we do not execute
    my $noexec = (defined($options{exec}) && (!$options{exec})) ? 1 : 0;

    # If verify is given, we ensure that the user wants to perform this command
    if ( ( $options{verify} ) && ( ! $noexec ) ) {
        print(STDERR "${indent}NOTE: Execute '$command'? ");
        my $ans = '';
        while ( $ans !~ m/^(y|n|yes|no)$/i ) {
            $ans = <STDIN>;
            chomp($ans);
            $ans = lc($ans);
        }
        if ( $ans =~ m/^(y|yes)$/ ) {
        } else {
            $noexec = 1;
        }
        $notedent .= '   ';
    }

    printf(STDERR "${indent}NOTE:%s%s %s\n", $notedent, ($noexec ? ' not really' : ''), $message) unless $options{quiet};
    printf(STDERR "$(indent}NOTE:${notedent}$command\n") if $verbose >= 3;

    my $result;
    if ( $noexec ) {
        $result = 1;
    } else {
        system($command);
        if ( $? != 0 ) {
            printf(STDERR "WARNING: $error (status %d = '%s')\n", $?, $!);
            die if $options{die};
        }
        $result = ! $?;
    }
    return $result;
}


sub mutateFile {
    # This function takes input and output file names/fps,
    # and copies lines from input to output while replacing
    # requested strings with corresponding values.
    #
    # $infile and $outfile are either strings representing
    # file names, or IO::Handle instances already opened
    # on files.  One can be a string while the other is
    # a IO::Handle, both can be strings, or both can be handles.
    #
    # @specs is an array of two-tuples, the first element of which
    # is a regexp to match, the second being its replacement.
    my ($infile, $outfile, @specs) = @_;

    my @errors;

    my $ifp = (ref($infile) && $infile->isa('IO::Handle')) ? $infile : IO::File->new("<$infile");
    if ( ! $ifp ) {
        push(@errors, "Failed to open '$infile' for reading.");
    }
    my $ofp = (ref($outfile) && $outfile->isa('IO::Handle')) ? $outfile : IO::File->new(">$outfile");
    if ( ! $ofp ) {
        push(@errors, "Failed to open '$outfile' for writing.");
    }

    if ( !@errors) {

        my $line = 0;
        foreach (<$ifp>) {
            ++$line;
            foreach my $tuple (@specs) {
                my ($match,$replace) = @$tuple;
                if ( m/$match/ ) {
                    s/$match/$replace/;
                }
            }
            print($ofp $_);
        }
    }

    ###
    # We close the file handles if we opened them here,
    # otherwise we leave them open for the caller to deal with
    $ifp->close() unless ref($infile);
    $ofp->close() unless ref($outfile);

    return (@errors) ? 0 : 1;
}


#########:
# Purpose: Print myself on the given filehandle (STDOUT by default)
# Args   :
# Notes  :
# Return :

sub print
{
    my $var = shift;
    my $fp = shift;
    my %format = @_;
    $fp = \*STDOUT unless defined($fp);

    print $fp &print_string( $var, %format );
}

#########:
# Purpose: Test all methods implemented in this module
# Args   :
# Notes  :
# Return :

sub _test_print
{
    my $obj = {};

    my @data2 = (5, 6, 8);
    my %data = ( 'a' => \@data2, 'b' => 'blork');
    my @data = (1,\%data,3,4);

    use vars qw($fpref);
    local $Wmh::fpref = \*OUT;
    open($$fpref,">blah");

    $obj->{'constant'} = \3.1415;
    $obj->{'ref'} = \\'hello';
    $obj->{'ref2'} = \\%data;
    $obj->{'function'} = \&func;
    $obj->{'file'} = $fpref;
    $obj->{'scalar'} = 2.1819;
    $obj->{'array'} = \@data;

    &print($obj, \*STDOUT,
           keyform => '%-20s',
           idxdent => '     '
           );

    close(OUT);
}


# Returns number of files in given directory
sub file_count {
    my ($dir) = @_;
    my $result;
    if ( ! -d $dir ) {
        $result = -1;
    } else {
        $result = (split(/\s+/, `cd $dir; ls | wc`))[1];
    }
    return $result;
}

sub makedir {
    # Make a directory safely
    my ($dir) = @_;
    #print("CHECKING $dir\n");
    my $result = 1;
    if ( ! -d $dir ) {
        if ( -e $dir ) {
            $result = 0;
            &croak("ERROR: Attempt to make directory '$dir' when non-directory version already exists.");
        } else {
            #print("   MAKING $dir\n");
            if ( system("mkdir -p \"$dir\"") ) {
                $result = 0;
                die("ERROR: Failed to create directory $dir");
            }
        }
    } else {
        $result = 1;
    }
    return $result;
}
sub symlink_old {
    # Make a symbolic link from $real to $alias
    #   If $alias exists and is a sym-link, it is removed.
    #   If $alias exists and is not a sym-link, an error is generated.
    #   If $real does not exist in $verify is true (default), an error is generated.

    my ($real, $alias, $verify) = @_;
    $verify = 1 unless defined($verify);

    if ( -l $alias ) {
        unlink $alias;
    } elsif ( -e $alias ) {
        &croak("ERROR: Attempt to make symbolic link but new file '$alias' already exists.");
    }

    if ( ( $verify ) && ( ! -e $real ) ) {
        &croak("ERROR: Attempt to make symbolic link from non-existing file $real.");
    } else {
        system("ln -s $real $alias");
    }
}


####################### Private Functions ##########################

#########:
# Purpose:
# Args   :
# Notes  :
#    Print out given value, no matter what type it is.
#
#    All lines are indented by prepending each with '$indent'
#    (which is usually a string of blanks).
#
#    If the maximum recursion depth has been reached, the string
#    sequence '...' is printed.
#
# Return :

sub _print_string
{
    my($var,$indent,$formref,$maxdepth) = @_;
    $indent = '' unless $indent;
    $var = 'undefined' unless defined($var);
    my $ref = ref($var);
    my $str = '';

    my $indent_length = length($indent);

    --$maxdepth;

    if ( $maxdepth == 0 ) {
        $str = "...\n";
    } elsif ( ! $ref ) {
        ++$ElementCount;
#       $str = &_annotateString($var,$indent_length,"E$ElementCount",$RightMargin) . "\n";
        # Don't bother annotating simple values
        $str = "$var\n";
    } else {
        my $type;
        my($prefix,$suffix) = ("\n","");

        if ( $var =~ m/^([^=]+)=([A-Z]+)\((.*)\)/ ) {
            # Class definition
            $indent = "$indent$formref->{'classdent'}";
            $str = "CLASS $1 ($3):\n$indent";
            $ref = $2;
        }

        ++$ElementCount;

        if ( defined( $ElementHash->{$var} ) ) {
            $str = "   ----> E$ElementHash->{$var}\n";
            $ref = '';  # so that it isn't processed below.
        } else {
            $ElementHash->{$var} = $ElementCount;
        }

        my $key;
        if ( $ref eq 'HASH' ) {
            $str .= &_annotateString($ref,$indent_length,"E$ElementCount",$RightMargin) . $prefix;

            # NOTE: Getting _annotate to work properly for the keys will be
            #       tricky - annotate assumes a single line, but for hashes
            #       and arrays, we need to annotate the first line of the
            #       return value of a recursive _print_string call.
            #       Wmh!!! Still to be implemented.

            foreach $key (sort(keys(%$var))) {
                $str .= sprintf("$indent$formref->{'keydent'}$formref->{'keyform'} = %s",
                                "'$key'",
                                &_print_string( $var->{$key},
                                               "$indent$formref->{'keydent'}",
                                               $formref, $maxdepth ) 
                               );
            }
            $str .= "$suffix";
        } elsif ( $ref eq 'ARRAY' ) {
            $str .= &_annotateString($ref,$indent_length,"E$ElementCount",$RightMargin) . $prefix;
            foreach $key (0..$#$var) {
                $str .= sprintf("$indent$formref->{'idxdent'}$formref->{'idxform'} = %s",
                                $key,
                                &_print_string( $var->[$key],
                                               "$indent$formref->{'idxdent'}",
                                               $formref, $maxdepth ) );
            }
            $str .= "$suffix";
        } elsif ( $ref eq 'SCALAR' ) {
            $str .= &_annotateString($ref . " $$var",$indent_length,"E$ElementCount",$RightMargin) . $prefix;
        } elsif ( $ref eq 'REF' ) {
            $str = &_print_string( "REF $$var", $indent, $formref, $maxdepth );
        } elsif ( $ref eq 'CODE' ) {
            $str = &_annotateString("FUNCTION $var",$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
        } elsif ( $ref eq 'GLOB' ) {
            my $fileno = fileno($var);
            my $s = 'FileHandle ';
            if ( defined($fileno) ) {
                $s .= "opened on fileno=$fileno";
            } else {
                $s .= 'unopened';
            }
            $str .= &_annotateString("$s",$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
        } else {
            # $ref does not match any of the known 'built-in' types,
            # and does not look like 'CLASS=BUILT-IN', so it must be
            # a "special" class with its own print format, so we just
            # print it verbatim.
            $str = &_annotateString($var,$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
        }
    }
    return $str;
}


#########:
# Purpose: Annotate a string with another string at the right margin
# Args   :
# Notes  :
#   This function takes a string, an indentation count, an annotation string
#   and a right margin index.  The string is printed, and the annotation
#   string is printed at the right margin.
#
# Example:
#    &_annotateString($var, 10, "Testing", 80);
# Return :

sub _annotateString
{
    my($result,$icnt, $annotate, $margin) = @_;

    my $len = $icnt + length($result);
    my $space = $margin - $len;
    $space = 1 unless $space > 0;

    $result .= sprintf("%-${space}s$annotate",' ');

    return $result;
}

#########:
# Purpose: Obtain
# Args   :
# Notes  :
# Return :

sub _generate_print_format
{
    my %format = @_;

    foreach (keys(%PrintFormat)) {
        $format{$_} = $PrintFormat{$_} unless $format{$_};
    }

    return(%format);
}

sub indent_string {
    &_indent_string(@_);
}

# Add $indent space before each line of given string.
# Wrap lines larger than $margin - $indent characters long
# If the first line should not be indented, use '+$indent' (i.e. '+4')
sub _indent_string {
    my ($str, $indent, $margin, $para_regexp, %options) = @_;
    # I SHOULD REALLY CLEAN UP THIS METHOD - USE %options AND EXPAND CAPABIITIES!
    # %options are:
    #    verbatim   --> if given, any line matching this regexp is not indented but is intead used verbatim (but indent is added)

    $indent = 4 unless defined($indent);

    $str =~ s/
/ /g;
    # If the $indent has an initial '+', do NOT indent first line
    # Note that $indent has the '+' removed if it existed.
    my $indentFirst = ($indent =~ s/^\+//) ? 0 : 1;

    # The user can specify $indent as a number (number of spaces to indent) or
    # as a literal string.  If it is a string, change $indent to be the length
    # of that string!
    my $space;
    if ( $indent =~ m/^\d+/ ) {
        # Number
        $space =  ' ' x $indent;
    } else {
        $space = $indent;
        $indent = length($indent);
    }

    # At this point, $indent should be a number!
    my $result;

    ###
    # Now deal with verbatim stuff
    my $verbatim = $options{verbatim};
    my %verb;
    if ( $verbatim ) {
        my $count = 1;
        while ( $str =~ s/((?:\n$verbatim.*)+)/\n\nREPLACETHIS$count/ ) {
            $verb{$count} = $1;
            ++$count;
        }
    }

    if (defined($margin)) {

        # wrap at $margin - $indent
        my $length = $margin - $indent;
        $result = '';

        if ( 1 ) {
            # Parse each line of $str.  If any line starts with regexp
            # $para_regexp, start a new paragraph.  If the line does
            # not start a new paragraph, add it to the current
            # paragraph, making sure that individual lines do not
            # exceed length '$length'

            # We first separate the $str into paragraphs.  We then
            # format each paragraph separately.

            my $para_regexp ||= qr/\n\s*\n/;
            my @paras = split(/$para_regexp/, $str);

            foreach my $para (@paras) {
                my $p = '';
                (my $initial = $para) =~ s/\s*\n/ /g;
                my $rest;

                ($initial, $rest) = &_split_str($initial, $length);
                my $predent = $space;
                while ( $rest ) {
                    #print "REST ($length): $rest\n";
                    # We have an initial segment in $initial that we add to $para,
                    # keeping track
                    $p .= $predent . $initial . "\n";
                    ($initial,$rest) = &_split_str($rest, $length);
                    $predent = $space;
                }
                $p .= $predent . $initial . "\n";
                $result .= "$p\n"
            }

        } elsif ( 0 ) {
            # Parse each line of $str.  If any line starts with regexp
            # $para_regexp, start a new paragraph.  If the line does
            # not start a new paragraph, add it to the current
            # paragraph, making sure that individual lines do not
            # exceed length '$length'

            my $para = '';
            my $line = '';
            my $indented;
            my $cnt = 0;
            my @lines = (split(/\n/, $str), '');
            my $N = $#lines + 1;
            my $z = 0;
            my $len;
            my $empty;

            my $initial = '';
            while ( $z < $N ) {
                $_ = $initial . $lines[$z++];

                $empty = ! m/\S/;

                if ( ( $empty ) ||
                     ( ( $para_regexp ) && ( m/^$para_regexp/ ) ) ) {

                    # Add previous paragraph to result
                    $para .= $space . $line . "\n"; # . ( $empty ? "\n" : '');
                    $result .= $para unless $para !~ m/\S/;
                    print(STDERR "NOTE: Added paragraph '$para'\n");

                    # Start a new paragraph.
                    $para = '';
                    $line = '';
                    $cnt = 0;

                    next if $empty;
                }

                # Add $_ to $line, then obtain a paragraph from it by wrapping
                # at margin $length.
                ($initial,my $rest) = &_split_str($initial, $length);
                my $predent = (($z==1)&&($indentFirst==0)) ? '' : $space;
                while ( $rest ) {
                    # We have an initial segment in $initial that we add to $para,
                    # keeping track
                    $para .= $predent . $initial . "\n";
                    ($initial,$rest) = &_split_str($rest, $length);
                    $predent = $space;
                }
                # $initial is now the last portion of what it was before the loop.
                # We add the next line
            }

        } else {
            ######
            # This is OLD code that did not handle $para_regexp!!!
            #####

            # Wrap each paragraph separately
            foreach (split(/\n\s*\n/, $str)) {
                s/\n/ /g; # make one big string with no newlines
                my $cnt = 0;
                while ( m/^(.{$length}\S*)\s*(.*)/ ) {
                    # Add first part of line to result, after indentation (but don't add
                    # any indentation to the first line if $indentFirst is 0)
                    $result .= ((($cnt==0)&&($indentFirst==0)) ? '' : $space) . $1 . "\n";
                    $_ = $2;
                    ++$cnt;
                }
                # add last bit of text if it isn't all whitespace
                $result .= ((($cnt==0)&&($indentFirst==0)) ? '' : $space) . $_ . "\n" if $_ !~ m/^\s*$/;
                # add a newline to separate paragraphs
                $result .= "\n";
            }
        }

        chomp($result);  # remove last \n
        if ( $result eq '' ) { $result = "\n"; } # for consistency

    } else {
        # No margin calculation is necessary, so things are easy!

        $result = $str;
        $result =~ s/\n/\n$space/g;
        $result = "$space$result" if $indentFirst;
    }


    ###
    # Now deal with the verbatum stuff again.
    foreach my $key (keys(%verb)) {
        my $v = $verb{$key};
        $v =~ s/^\s*\n//;
        $v =~ s/\n/\n$space/g;
        $result =~ s/REPLACETHIS$key/$v/;
    }

    return $result;
}

sub _split_str {
    # Returns array of two strings obtained by splitting $str
    # at the closest whitespace beyond index $index.  If the
    # split position is beyond the length of $str, the first
    # returned string is the entire input argument, and the
    # second one is empy ('').
    #
    # Args:
    #   [0]  = string
    #   [1]  = index to split at

    my $i = $_[1];
    my $l = length($_[0]);

    if ( $_[0] =~ m/^(.{0,$i}\S*)\s*(.*)/ ) {
        my ($h,$t) = ($1,$2);
        my $l = length($h);
        if ( $l > $i ) {
            # We overflowed by one word - remove from $h and preprend to $t
            $h =~ s/(\S+)\s*$//;
            my $p = $1;
            if ( $h eq '' ) {
                # $h consisted of a single non-whitespace word
                # that exceeds our length request.  We will
                # get an infinite loop if we don't consume
                # something - there is no way to indent this
                # line within the limits specified, so we just
                # bite the bullet.
                $h = $p;
            } else {
                $t = "$1 $t";
            }
        }
        return ($h,$t);
    } else {
        die("ERROR: Couldn't split '$_[0]' at $i\n");
    }
}

sub marginize_str {
    # Given a string (without newlines) and a margin, returns a wrapped
    # string (with newlines).
    my ($str, $margin) = @_;
    my $res = '';
    my ($initial,$rest) = &_split_str($str, $margin);
    while ( ! $rest ) {
        $res .= $initial;
        ($initial,$rest) = &_split_str($rest, $margin);
    }
    $res .= $initial;
    return $res;
}

# I'm always running across a need to compute various statistical values
# based on a list of numbers.  Here is the code to do so, all in one simple
# routine.  Returns a reference to a hash whose keys are stastical concepts
# and whose values are the value of the concept for the input list of numbers
sub statistics(@) {
    my %result;

    # Compute necessary variables by iterating over each element of input list
    my $N = scalar(@_);
    my ($sum, $sumsq) = (0,0);
    my ($max, $min) = (0,100000);
    my %occurences;
    foreach (@_) {
        $sum += $_;
        $sumsq += $_ ** 2;
        $max = $_ if ($_ > $max);
        $min = $_ if ($_ < $min);
        $occurences{$_} = 0 unless defined($occurences{$_});
        ++$occurences{$_};
    }

    # Need to search %occurences for the key with the largest value
    my $cnt = 0;
    my $mode = '';
    foreach (keys(%occurences)) {
        if ( $occurences{$_} > $cnt ) {
            $cnt = $occurences{$_};
            $mode = $_;
        }
    }

    # The median is the middle number (or average of middle two if even number of
    # elements)
    my $mid;
    my $median;
    if ( ($N % 2) == 0 ) {
        # Even number of elements
        $mid = ( $N / 2 ) - 1;
        $median = ($_[$mid] + $_[$mid+1]) / 2;
    } else {
        # Odd number of elements
        $mid = ($N - 1) / 2;
        $median = $_[$mid];
    }

    # Store the results
    $result{'max'} = $max;
    $result{'min'} = $min;
    $result{'mode'} = $mode;
    $result{'median'} = $median;
    $result{'mean'} = $sum / $N;
    $result{'variance'} = (($N * $sumsq) - ($sum ** 2)) / ( $N * ($N - 1) );
    $result{'stddev'} = sqrt($result{'variance'});
    $result{'coeffvar'} = ( $result{'stddev'} / $result{'mean'} ) * 100;

    return \%result;
}

sub capitalizeWords {
    $_[0] =~ s/([^\sA-Za-z]*)([a-zA-Z])(\S*)(\s+|$)/$1\u$2$3$4/g;
    return $_[0];
}

sub centerStr {
    # Returns a string padded on left and right by an appropriate amount so that
    # it is "centered" within $width units.
    my ($str, $width, $delim) = @_;
    $delim ||= ' ';
    my $len = length($str);
    if ( $len >= $width ) { return $str };
    my $head = int(($width - $len)/2);
    my $tail = $width - $len - $head;

    return ($delim x $head) . $str . ($delim x $tail);
}

sub wsprint {
    # Given a string (with embdedded substition requests) and a list of tag/value
    # substition specifications, return a modified string by making all
    # requested substitions.
    #
    # An optional argument '$obj' can be specified.  If tags found
    # in $format are not defined in $state, the code checks to see if a method for
    # $tag exists in $obj (if one does, we use its return value as the value to
    # substitute, instead of exiting with error).
    #
    # If specified, $tagformat describes the format of each tag.  It should
    # have a single grouping specification that grabs everything inside the
    # tag delimiters.  If not provided, a default specification is used that
    # matches against tags that look like this: <=tag>
    #
    # When substituting tags for values, any trailing newlines on $value are
    # removed before the substitution occurs.  This ensures that the template
    # format '$format' looks similar to the actual output.

    my ($format, $state, $obj, $tagformat) = @_;
    $tagformat = '<=([^>\s]+(?:@[^>]+)?)>' unless defined($tagformat);

    ### NOTE: (!!!WMH) The algorithm for doing this substitution should be
    ### changed - as it is now, the substituted text may have sequences
    ### that look like $tagformat (even though the user does not want any
    ### substitutions to occur for such text).

    my @lines;
    foreach (split(/\n/, $format)) {
        # In order to provide the readability of perl source code, it is
        # nice to indent the lines describing generated text, but not have
        # such indentation show up in the generated output.  In order to
        # provide this, any line starting with arbitrary whitespace followed
        # by a '|' is stripped off.
        s/^\s*\|//;

        # Look for tags on the line
        while ( s/$tagformat/NEXTSUB/ ) {

            my $replace = $1;
            if ( $replace =~ m/([^@]+)(?:@(.*))?/ ) {
                my ($tag,$value,$sprint) = ($1,$state->{$1},$2);

                if ( ! defined($value) ) {
                    # Check to see if there is a method defined on $self with
                    # the name '$tag', and if so, assign both $state->{$tag} and
                    # $value to the result of calling that function.  Otherwise,
                    # exit with an error message

                    if ( defined($obj) ) {
                        eval("\$value = \$obj->$tag();");
                        if ( $@ ) {
                            &carp("Could not invoke method '$tag' on given object: $@\n");
                        } else {
                            $value = 'UNDEF' unless defined($value);
                            $state->{$tag} = $value;
                        }
                    } else {
                        # Error: no value for tag $tag exists in $state
                        &carp("No value for tag '$tag' specified in state!");
                    }
                }

                # The tag has an associated value in $state - do substitution
                if ( defined($sprint) ) {
                    # We have an array specification, so the tag is replaced
                    # with the concatenation of the strings formed by evaling
                    # sprintf($sprint, $val) for each element $val is @{$value}

                    my $ref = ref($value);
                    &carp("ERROR: tag $tag should have an ARRAY value (not type '$ref')") && exit(1)
                      if ( $ref ne 'ARRAY' );

                    $sprint =~ s/\\n/\n/g;
                    $sprint =~ s/\\t/\t/g;

                    my $subst = '';
                    my $v;
                    foreach $v (@$value) {
                        croak("Undef value for $tag") if ! defined $v;
                        $subst .= sprintf($sprint, $v) . "\n";
                    }

                    $subst =~ s/\n+$//;
                    s/NEXTSUB/$subst/;

                } else {
                    # We have a simple scalar substitution request - replace
                    # the entire tag construct (NEXTSUB) with the value $value
                    $value =~ s/\n+$//;
                    s/NEXTSUB/$value/;
                }
            }
        }
        push(@lines, $_);
    }
    return join("\n", @lines);
}

sub initHash {
    # Given a base HASH $base and a 'default-value' hash $defaults,
    # initialize fields of $base from $defaults if values don't
    # already exist.
    my ($base, $defaults,$keys, $forcekeys, $verbose) = @_;
    $verbose = 0 unless defined($verbose);

    my @keys = (defined($keys)) ? @$keys : keys(%{$defaults});
    my @forcekeys = (defined($keys)) ? @$forcekeys : ();

    print "Initializing hash from defaults...\n" if $verbose;
    foreach (@keys) {
        if ( !defined($defaults->{$_}) ) {
            &carp("   Warning: key '$_' not defined in defaults-array (ignored)")
                if $verbose;
        } else {
            if ( ! defined($base->{$_}) ) {
                printf("   Added %-15s => %s\n", $_, $defaults->{$_}) if $verbose;
                $base->{$_} = $defaults->{$_};
            }
        }
    }
    foreach (@forcekeys) {
        if ( !defined($defaults->{$_}) ) {
            &carp("   Warning: key '$_' not defined in defaults-array (ignored)")
                if $verbose;
        } else {
            printf("   Added  %-15s => %-20s  (overwrote %s)\n",
                   $_, $defaults->{$_}, (defined($base->{$_}) ? $base->{$_} : 'undef'))
                if $verbose;
            $base->{$_} = $defaults->{$_};
        }
    }
}

sub warghash(@) {
    # Process the @_ of the caller.  Returns a HASH reference, which comes
    # from either the first argument (if it is a HASH reference), or from
    # all arguments treated as a hash.
    my $result;
    if ( ( $#_ > -1 ) && ( ref($_[0]) eq 'HASH' ) ) {
        $result = $_[0];
    } else {
        my %result = @_;
        $result = \%result;
    }
    return $result;
}

sub findModulePaths {
    # Returns an array of source paths used for given module names
    my @res;
    foreach (@_) {
        push(@res, [findModulePath($_)]);
    }
    return @res;
}

sub findModulePath {
    # Return all paths of the .pm file used by perl for a given module name.
    # Empty array is returned if no such module exists in @INC.
    my ($modname, $dirOnly) = @_;
    $dirOnly ||= 0;

    $modname =~ s%::%/%g;
    my @res = ();
    foreach (@INC) {
        # print STDERR "WORKING ON '$_'\n";
        my $suff;
        foreach $suff (qw(pm pl)) {
            if ( -f "$_/$modname.$suff" ) {
                push(@res, (( $dirOnly ) ? $_ : "$_/$modname.$suff"));
            }
        }
    }
    return ( wantarray ) ? @res : (( $#res > -1 ) ? $res[0] : undef);
}

sub open_safe {
    # Return an instance of IO::File opened on the given specification (as
    # accepted by the 'open' builtin.
    #
    # We keep the signature the same as for 'open', except that another
    # optional $opt_err arg is allowed, specifying what error message
    # should be printed out.  If the file cannot be opened, the program
    # will exit immediately with a meaningful error message.

    my ($arg1, $arg2, $opt_err) = @_;

    # This conditional call is necessary because the default code does
    # tested passed on the number of arguments, not on whether $arg2
    # is defined or not

    my $res = ( defined($arg2) ) ?
        IO::File->new($arg1,$arg2) :
        IO::File->new($arg1);
    if ( ! $res ) {
        # Based on the mode, generate an error message to print out
        my $err = $opt_err;
        if ( ! defined($arg2) ) {
            # We are in the normal perl version in which both file and
            # mode are encoded in $arg1
            my $mode = $arg1;

            if ( $mode =~ m/^\s*(>{1,2})\s*(\S+)/ ) {
                # Writing or appending
                $err = sprintf("Failed to open '$2' for %s\n", ($1 eq '>') ?
                               'writing' : 'appending');

            } elsif ( $mode =~ m/^\s*<?\s*(\S+)/ ) {
                # Reading
                $err = "Failed to open '$1' for reading";

            } elsif ( $mode =~ m/^\s*\|\s*(.*\S)/ ) {
                # Piping out
                $err = "Failed to open a pipe out to command '$1'";

            } elsif ( $mode =~ m/^\s*(.*)\|\s*$/ ) {
                # Piping in
                $err = "Failed to open a pipe in from command '$1'";

            }
        } else {
            # We are in the C version where $arg1 is a file name and
            # $arg2 specifies how the file should be opened.
            my ($file, $mode) = ($arg1, $arg2);

            $err = "Failed to open '$file' (mode '$mode' has not be englified yet)";
        }

        carp("ERROR: $err (\L$!\E)");
        exit(1);
    }

    return $res;
}

sub readTo {
    # Return a string representing all lines in $fp up to the line
    # matching $regexp.  If $opt_first is true, initialize result
    # to $_ (before reading any lines).  If $opt_end is true,
    # add the matching termination line.
    #
    # Returns a multi-line string if evaluated in scalar context, and
    # an array reference of strings (WITHOUT newlines) in a list context.

    my ($fp, $regexp, $opt_start, $opt_end) = @_;
    my @result;

    # Determine if the current line should be added to the result
    if ( $opt_start ) {
        chomp($_);
        push(@result, $_);
    }

    # Scan the file until the regexp is found
    while (<$fp>) {
        chomp;
        last if m/$regexp/;
        push(@result, $_);
    }

    # Determine if the regexp line should be added to the result
    push(@result, $_) if $opt_end;

    # Return a string in scalar context, list in list context
    return wantarray ? \@result : join("\n", @result, '');
}


sub time_difference {
   return (stat($_[0]))[9] - (stat($_[1]))[9];
}

sub parseKeywordValues {
    # Return a hash from a string of keyword/value pairs
    my ($str, $delim) = @_;
    $delim ||= '=';

    $str = " $str";
    my $res = {};
    my @res;
    while ( $str =~ s/(.*)\s+([^\s$delim]+)\s*$delim\s*(.*\S).*/$1/ ) {
        push(@res, [$2,$3]);
    }
    foreach (reverse(@res)) {
        $res->{$_->[0]} = $_->[1];
    }
    return $res;
}

sub parseSpecialKeywordValues {
    # Allows caller to specify the key, delimiter and value formats,
    # as well as a filter rule to apply to values.  For example, in
    # order to extend the normal semantics provided by 'parseKeywordValues'
    # to allow values to be empty strings, do the following:
    #    Wmh::parseSpecialKeywordValues($str,
    #                                   value => qr/.*/,
    #                                   filter => qr/(.*?)\s*$/
    #                                   );

    my ($str,%opts) = @_;
    my $keyre   = $opts{key}   || qr/[^\s=]+/;
    my $delimre = $opts{delim} || qr/\s*=\s*/;
    my $valre   = $opts{value} || qr/.*\S/;
    my $filter  = $opts{filter};  # if specified, should have one group - the filter is applied to each value, and the value is changed to be the result of the subgroup match.

    $str = " $str";
    my $res = {};

    while ( $str =~ s/(.*)\s+($keyre)$delimre($valre)/$1/ ) {
        my ($k,$v) = ($2,$3);
        if ( $filter ) {
            $v =~ s/$filter/$1/;
        }
        $res->{$k} = $v;
    }
    return $res;
}

sub joinKeywordValues {
    my ($hash,@porder) = @_;
    my @order;
    my %order;
    foreach my $k (@porder) {
        push(@order, $k);
        $order{$k} = 1;
    }
    foreach my $k (sort keys(%$hash)) {
        if ( ! $order{$k} ) {
            push(@order,$k);
        }
    }

    my @res;
    foreach (@order) {
        my $v = $hash->{$_};
        if ( defined($v) ) {
            push(@res, "$_=" . $hash->{$_});
        }
    }
    return join(' ', @res);
}

sub fixedLength {
    # Returns the first $len characters of $str (or the whole string
    # if its length is less then $len
    my ($str, $len) = @_;
    if ( length($str) <= $len ) {
        return $str;
    } else {
        my $l = $len - 1;
        $str =~ m/^(.{0,$l})/;
        return $1 . "\$";
    }
}

sub ask {
    my ($question, $fp) = @_;
    $fp ||= \*STDOUT;
    my $ans;
    # autoflush $fp 1;
    while (1) {
        print($fp "$question? ");
        $ans = <STDIN>;
        last if ( $ans =~ m/^y|n/i );
    }
    return ( $ans =~ m/^y/i );
}

sub askdef {
    my ($question, $default, $fp) = @_;
    $fp ||= \*STDOUT;
    my $ans;
    # autoflush $fp 1;
    my $prefix = "$question";
    if ( defined($default) ) {
        $prefix .= " [$default]";
    }
    $prefix .= "? ";
    while (1) {
        print($fp $prefix);
        $ans = <STDIN>;
        chomp($ans);
        if ( defined($default) ) {
            $ans ||= $default
        }
        last if ( $ans =~ m/^y|n/i );
    }
    return ( $ans =~ m/^y/i );
}

sub askForValue {
    my ($question, $re, $default) = @_;
    $re ||= qr/.*/;
    $default ||= '';

    my $fullquestion = $question;
    if ( $default ) {
        $fullquestion .= " ($default)";
    }

    my $ans;
    autoflush STDOUT 1;
    while (1) {
        print(STDOUT "$fullquestion? ");
        $ans = <STDIN>;
        chomp($ans);
        $ans ||= $default if $default;
        last if ( $ans =~ m/$re/ );
    }
    return $ans;
}

sub wmh_rangeOrList {
    # Returns an array of integers obtained from a string.
    # The string allows integer ranges and lists of integers
    # to be intermixed.  The only legal characters in the
    # input stream are '-', ',', whitespace and digits.
    #
    # If $min is given, it indicates the smallest legal integer
    # (usually used when the $spec is something like '-7').
    # If $max is given, it indicates the largest legal integer
    # (usually used when the $spec is something like '7-').
    #
    # Returns undef if the specification was illegal

    my ($spec, $min, $max) = @_;

    # Remove whitespace
    $spec =~ s/\s+//g;
    my @prelim = split(/,/, $spec);
    my @list;

    my @errors;

    foreach (@prelim) {
        if ( m/^\d+$/ ) {
            push(@list, $_);

        } elsif ( m/^-(\d+)$/ ) {
            if ( defined($min) ) {
                push(@list, $min..$1);
            } else {
                push(@errors, "No minimum specified but format '$_' requires one");
            }

        } elsif ( m/^(\d+)-$/ ) {
            if ( defined($max) ) {
                push(@list, $1..$max);
            } else {
                push(@errors, "No maximum specified but format '$_' requires one");
            }

        } elsif ( m/^(\d+)-(\d+)$/ ) {
            if ( $1 < $2 ) {
                push(@list, $1..$2);
            } else {
                push(@errors, "Range '$_' has min > max");
            }
        } else {
            # Invalid format
            push(@errors, "Illegal field format '$_'");
        }
    }

    if ( @errors ) {
        print join("\n", @errors), "\n";
        return undef;
    } else {
        my @result = sort { $a <=> $b } @list;
        return (wantarray) ? @result : \@result;
    }
}

sub loop {
    my ($loopvars) = @_;
    # $loopvars is an array of arrays representing the values
    # to loop over in each dimension.
    #
    # Returns an array of arrays (each subarray has N
    # elements, where N is the number of dimensions in the
    # loop, and the value in index i is a value for
    # $loopvars[i]).

    my $results = [];
    &_loop($loopvars, [], $results);
    return $results;
}

sub _loop {
    my ($loopvars, $indices, $results) = @_;

    ###
    # $loopvars is an array of arrays representing the values
    # to loop over in each dimension.
    #
    # $indices establishes the indexes within the loop array
    # of the "current" iteration.
    #
    # $results is an array of arrays (each subarray consists
    # of a particular set of values in $loopvars.
    #
    # For example:
    #   &loop( [[1,2,3], [5, 7], [1,2]], [], [] )
    # returns:
    #   [ [1,5,1], [1,5,2], [1,7,1], [1,7,2],
    #     [2,5,1], [2,5,2], [2,7,1], [2,7,2],
    #     [3,5,1], [3,5,2], [3,7,1], [3,7,2],
    #   ]
    #

    # $i indicates which dimension we are in
    my $i = $#$indices;

    # if $i == $#$loopvars, we are at the lowest dimension, and the
    # values in $indices represents indices into the arrays in $loopvars
    if ( $i == $#$loopvars ) {
        my @tuple = map { $loopvars->[$_]->[ $indices->[$_] ] } (0..$i);
        push(@$results, [@tuple]);
        # print( ': ' . join(' ', @$indices) . '     ' . join(' ', @tuple) . "\n");
    } else {
        # We are not at the lowest dimension yet, so we loop over all values
        # in the next dimension, recursively invoking the routine with
        # a modified $indices array.
        # print "\n";
        my $dim = $loopvars->[$i+1];
        for my $loop (0..$#$dim) {
            # printf("%sLOOP %d", ('  ' x ($i+1)), $loop);
            $indices->[$i+1] = $loop;
            &_loop($loopvars, $indices, $results);
        }
        pop(@$indices);
    }
}

############################# Raw Code  ############################
				      
				      
########################## Class Methods ###########################



########################## Instance Methods ########################

###
# Accessors

# End Accessors
###

###
# User Code
#line 1863 "wmh.pxx"
sub addData {
    carp('Usage: Wmh::addData(data, key, val)') unless (@_ == 3);
    my($data) = $_[0]; 
    my($key) = $_[1]; CORE::ref($key) eq '' or carp(sprintf('Wmh::addData: ref($key) = %s instead of .', CORE::ref($key))); 
    my($val) = $_[2]; 
#line 1864 "wmh.pxx"
    my $ref = ref($data);
    my $valref = ref($val);

    # print "ADDING $key = $val to $data\n";

    if ( $ref eq 'ARRAY' ) {
        if ( ! $valref ) {
            # If the current value is defined, we append.
            $data->[$key] = ($data->[$key] || '') . $val;
        } else {
            $data->[$key] = $val;
        }
    } elsif ( $ref eq 'HASH' ) {
        if ( ! $valref ) {
            $data->{$key} = ($data->{$key} || '') . $val;
        } else {
            $data->{$key} = $val;
        }
    } else {
        printf(STDERR "WARNING: Unknown data type '$ref' in addData\n");
    }
}

#line 1608 "wmh.pxx"
sub canonicalPhoneNumber {
    carp('Usage: Wmh::canonicalPhoneNumber(phone_in, default_area)') unless ( (@_ >= 1) && (@_ <= 2) );
    my($phone_in) = $_[0]; CORE::ref($phone_in) eq '' or carp(sprintf('Wmh::canonicalPhoneNumber: ref($phone_in) = %s instead of .', CORE::ref($phone_in))); 
    my($default_area) = $_[1]; $default_area = '' unless defined($default_area); CORE::ref($default_area) eq '' or carp(sprintf('Wmh::canonicalPhoneNumber: ref($default_area) = %s instead of .', CORE::ref($default_area))); 
#line 1609 "wmh.pxx"
    my $phone = $phone_in;


    my $result;

    ###
    # Extract extension
    my $extension;
    if ( $phone =~ s%(?:ext.?|x)\s*(\d+)%%i ) {
        $extension = $1;
    }

    ###
    # Extract annotation (i.e. 'home', 'office', 'cell', etc.)
    my $annotation = '';
    if ( $phone =~ s/^\s*([a-zA-Z]+)\s*:\s*// ) {
        $annotation = $1;
    } elsif ( $phone =~ s/\s*\([a-zA-Z\s]+\)\s*// ) {
        $annotation = $1;
    }
    
    ###
    # Parse out long-distance indicator
    my $longdist;
    if ( $phone =~ s/^\s*(1)[\s.-]+// ) {
        $longdist = $1;
    }
    
    ###
    # Parse out the area, prefix and suffix
    my ($area,$prefix,$suffix);
    if ( $phone =~ m/^\s*(?:\((\d{3})\)\s*[.-]?\s*)?(\d{3})\s*[.-]?\s*(\d{4})\s*$/ ) {
        ($area,$prefix,$suffix) = ($1,$2,$3);
    } elsif ( $phone =~ m/^\s*(?:(\d{3})?\s*[.-]?\s*)(\d{3})\s*[.-]?\s*(\d{4})\s*$/ ) {
        ($area,$prefix,$suffix) = ($1,$2,$3);
    } else {
        # Bail - could not find correct version.
        if ( $phone_in ) {
            print(STDERR "WARNING: Number '$phone_in' not canonical\n");
            $result = undef;
        } else {
            $result = $phone_in;
        }
        return $result;
    }
    $area ||= $default_area;

    ###
    # Obtain the result;
    if ( ! defined($result) ) {
        if ( $area ) { $area = sprintf("%3d-", $area); }
        $result = sprintf("%s%s%03d-%04d%s",
                          (($annotation) ? "$annotation:" : ''),
                          $area,
                          $prefix,
                          $suffix,
                          (($extension) ? "x$extension" : ''),
                         );
        # printf(STDERR "IN: %-20s  RESULT: $result\n", $phone_in);
    }
    return $result;
}

#line 1672 "wmh.pxx"
sub canonicalPhoneNumbers {
    carp('Usage: Wmh::canonicalPhoneNumbers(phonelist, default_area)') unless ( (@_ >= 1) && (@_ <= 2) );
    my($phonelist) = $_[0]; CORE::ref($phonelist) eq '' or carp(sprintf('Wmh::canonicalPhoneNumbers: ref($phonelist) = %s instead of .', CORE::ref($phonelist))); 
    my($default_area) = $_[1]; $default_area = '' unless defined($default_area); CORE::ref($default_area) eq '' or carp(sprintf('Wmh::canonicalPhoneNumbers: ref($default_area) = %s instead of .', CORE::ref($default_area))); 
#line 1675 "wmh.pxx"

    my @list;
    if ( $phonelist =~ m%[,/&]% ) {
        @list = split(/\s*[,\/]\s*/, $phonelist);
    } elsif ( $phonelist =~ m%\s+or\s+%) {
        @list = split(/\s+or\s+/, $phonelist);
    # } elsif ( $phonelist =~ m/\d\s+\(\d{3}\)/ ) {
        # Looks like two or more phone numbers with no separator
        #while ( $phonelist =~ s/(.*)(\([^\(])/$1/ ) {
        #push(@list, $2);
        #}
    } else {
        # Assume it represents one number
        @list = ($phonelist);
    }

    my @result;
    foreach my $num (@list) {
        my $cannum = &canonicalPhoneNumber($num, $default_area);
        push(@result, $cannum) if $cannum;
    }
    
    return (wantarray) ? @result : join(', ', @result);
}

#line 1887 "wmh.pxx"
sub dataByName {
    carp('Usage: Wmh::dataByName(name)') unless (@_ == 1);
    my($name) = $_[0]; CORE::ref($name) eq '' or carp(sprintf('Wmh::dataByName: ref($name) = %s instead of .', CORE::ref($name))); 
#line 1889 "wmh.pxx"
    my $result;
    $name = "\U$name\E";
    if ( $name eq 'ARRAY' ) {
        $result = [];
    } elsif ( $name eq 'HASH' ) {
        $result = {};
    } else {
        $result = undef;
    }
    return $result;
}

#line 1775 "wmh.pxx"
sub dataFromTextArray {
    carp('Usage: Wmh::dataFromTextArray(data, text, receiver)') unless (@_ == 3);
    my($data) = $_[0]; 
    my($text) = $_[1]; ((!$text) || (CORE::ref($text) eq 'ARRAY')) or carp(sprintf('Wmh::dataFromTextArray: ref($text) = %s instead of ARRAY.', ref($text))); 
    my($receiver) = $_[2]; 
#line 1795 "wmh.pxx"

    my @text = @$text;
    while ( @text > 0 ) {

        my $line = shift(@text);
        next if ( $line =~ m/^\s*$/ );

        if ( $line =~ m/^(\s*)(\S+)\s*=\s*(\S+)\s*(.*)/ ) {
            my ($indent, $key, $val1, $val2) = ($1,$2,$3,$4);
            $val2 =~ s/\s*$//;

            my $val;

            if ( ( $val1 eq 'ARRAY' ) || ($val1 eq 'HASH' ) ) {

                # The current value of the key with which the value will be associated
                my $cval = &getData( $data, $key );

                # Determine the actual value
                if ( defined($cval) ) {
                    if ( ref($cval) eq $val1 ) {
                        # Existing data-structure matches new data-structure, so we
                        # use the existing one.
                        $val = $cval;
                    } else {
                        # Existing data-structure does not match new one - report error
                        print(STDERR "ERROR: Attempt to add key '$key' of type '$val1' to data-structure that already has that key with type " . ref($cval) . "\n");
                        $val = undef;
                    }
                } else {
                    # Do not currently have a value for this key, so we create a new initial
                    # value of the appropriate data-structure type.
                    $val = ( $val1 eq 'ARRAY' ) ? [] : {};
                }

                ###
                # Now we obtain the sub-portions of @text corresonding to sub-information
                # of this data-structure.  All lines with more initial white-space the
                # the current line are consumed here.
                my @t; 
                while ((@text) &&
                       (($text[0] =~ m/^$indent\s+/ ) ||
                        ($text[0] =~ m/^\s*$/))) {
                    push(@t, shift(@text));
                }
                &dataFromTextArray($val, \@t, $receiver);

            } elsif ( $val1 eq 'TEXT' ) {
                my @t; 
                while ((@text) && ($text[0] =~ m/^$indent\s+/ )) {
                    my $line = shift(@text);
                    # $line =~ s/^\s*\|//;
                    push(@t, $line);
                }
                $val = "\n" . join("\n", @t);
            } elsif ( $val1 eq 'FUNCTION' ) {
                $val = &localFunction($val2,$receiver);
            } else {
                $val = $val1;
                $val .= ' ' . $val2 if $val2;
            }
            &addData($data, $key, $val);
        } else {
            print("WARNING: Unknown line ignored - $line\n");
        }
    }
}

#line 1901 "wmh.pxx"
sub getData {
    carp('Usage: Wmh::getData(data, key)') unless (@_ == 2);
    my($data) = $_[0]; 
    my($key) = $_[1]; CORE::ref($key) eq '' or carp(sprintf('Wmh::getData: ref($key) = %s instead of .', CORE::ref($key))); 
#line 1902 "wmh.pxx"
    my $ref = ref($data);
    my $result;
    if ( $ref eq 'ARRAY' ) {
        $result = $data->[$key];
    } elsif ( $ref eq 'HASH' ) {
        $result = $data->{$key};
    } else {
        printf(STDERR "WARNING: Unknown data type '$ref' in addData\n");
    }
    return $result;
}

#line 1 "GENERATED CODE"
sub initState {
    carp('Usage: Wmh::initState(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('Wmh'))) or carp(sprintf('Wmh::initState: NOT $self->isa(Wmh)')); 
#line 3 "GENERATED CODE"
{
    $_[0]->SUPER::initState();
    $_[0]->extend({
    });
}}

#line 1700 "wmh.pxx"
sub localFunction {
    carp('Usage: Wmh::localFunction(func, cgiclass)') unless (@_ == 2);
    my($func) = $_[0]; CORE::ref($func) eq '' or carp(sprintf('Wmh::localFunction: ref($func) = %s instead of .', CORE::ref($func))); 
    my($cgiclass) = $_[1]; 
#line 1705 "wmh.pxx"
    my $cls = $cgiclass;
    if ( ref($cgiclass) ) {
        $cls = $cgiclass->myclass();
    }

    my $hash;
    eval "\$hash = \\%${cls}::";
    my $result = $hash->{$func};
    if ( ! $result ) {
        my $parent;
        eval "\$parent = \$${cls}::ISA[0];";
        if ( $parent ) {
            $result = &localFunction($func, $parent);
        } else {
            printf(STDERR "WARNING: Could not find function '$func' for '$cls'\n");
        }
    }
    return $result;
}

#line 2046 "wmh.pxx"
sub readOneChar {
    carp('Usage: Wmh::readOneChar()') unless (@_ == 0);
#line 2053 "wmh.pxx"
    system('stty raw -echo') && die("Failed to set character mode");
    my $fp = \*STDIN;
    my $c;
    $c = $fp->getc();
    system('stty -raw echo')  && die("Failed to set line mode");
    return $c;
}

#line 2061 "wmh.pxx"
sub realPath {
    carp('Usage: Wmh::realPath(path)') unless (@_ == 1);
    my($path) = $_[0]; CORE::ref($path) eq '' or carp(sprintf('Wmh::realPath: ref($path) = %s instead of .', CORE::ref($path))); 
#line 2064 "wmh.pxx"
    my @path = split(/\//, $path);
    my @result;
    my $dir = '';
    if ( $path[0] eq '' ) {
        # Absolute path
        shift(@path);
    } else {
        $dir = shift(@path);
    }
    foreach my $node (@path) {
        $dir .= "/$node";
        #print "DIR: $dir\n";
        if ( -l $dir ) {
            my $link = readlink($dir);
            #print "  SYML $dir  LINK $link\n";
            if ( $link =~ m%^/% ) {
                # Full path - we replace $dir with $link
                $dir = $link;
            } else {
                # Relative path - somewhat more messy
                $dir =~ s%/[^/]*$%/$link%; # replace $node with $link
                # Now do special processing to handle '../' and './'
                $dir =~ s%([^/]+)/\.\./%/%g;
                $dir =~ s%/\./%%g;
            }
        }
    }
    #print("NOTE: '$path' becomes\n      '$dir'\n");
    return $dir;
}

#line 1914 "wmh.pxx"
sub relativeDirectory {
    carp('Usage: Wmh::relativeDirectory(src, dest)') unless (@_ == 2);
    my($src) = $_[0]; CORE::ref($src) eq '' or carp(sprintf('Wmh::relativeDirectory: ref($src) = %s instead of .', CORE::ref($src))); 
    my($dest) = $_[1]; CORE::ref($dest) eq '' or carp(sprintf('Wmh::relativeDirectory: ref($dest) = %s instead of .', CORE::ref($dest))); 
#line 1919 "wmh.pxx"
    
    $dest =~ s%/\s*$%%;  # remove trailing '/'
    my $tmp = "$src $dest";
    my $result;
    if ( $tmp =~ m%^(.*)/(.*) \1/(.*)% ) {
        my ($sb,$ssub,$dsub) = ($1,$2,$3);
        #print "RD: $tmp\n   SB  : $sb\n   SSUB: $ssub\n   DSUB: $dsub\n";
        my @ssub = split(/\//, $ssub);
        my $nrel = $#ssub; # num elements in @ssub - 1 (the parent directories)
        $result = join('/', (('..')x$nrel)) . '/' . $dsub;
    } else {
        $result = $dest;
    }
    #print "   REL : $result\n";

    return $result;
}

#line 1726 "wmh.pxx"
sub strToStruct {
    carp('Usage: Wmh::strToStruct(str, options)') unless (@_ >= 1);
    my($str) = $_[0]; CORE::ref($str) eq '' or carp(sprintf('Wmh::strToStruct: ref($str) = %s instead of .', CORE::ref($str))); 
    my(%options) = @_[1..$#_]; 
#line 1736 "wmh.pxx"

    if ( $str =~ m/^\s*$/ ) {
        # Need to deal with empty blocks.
        return {};
    }

    my (@text) = split(/\n/, $str);
    while (1) {
        if ( $text[0] =~ m/^\s*$/ ) {
            shift(@text);
        } else {
            last;
        }
    }

    my $result;
    if ( $text[0] =~ m/^(HASH|ARRAY)\s*$/ ) {
        my $base = shift(@text);
        if ( $base eq 'ARRAY' ) {
            $result = [];
        } elsif ( $base eq 'HASH' ) {
            $result = {};
        } else {
            carp("ERROR: Unknown base type '$base' in dataFromText\n");
            $result = undef;
        }
    } else {
        # Use the data-structure specified (assume hash if none given)
        $result = $options{data} || {};
    }
    map { s/^\s*\#.*//; } @text;
    
    if ( $result ) {
        &dataFromTextArray( $result, \@text, $options{receiver});
    }

    return $result;
}

#line 1937 "wmh.pxx"
sub symlink {
    carp('Usage: Wmh::symlink(src, dest, options)') unless (@_ >= 2);
    my($src) = $_[0]; CORE::ref($src) eq '' or carp(sprintf('Wmh::symlink: ref($src) = %s instead of .', CORE::ref($src))); 
    my($dest) = $_[1]; CORE::ref($dest) eq '' or carp(sprintf('Wmh::symlink: ref($dest) = %s instead of .', CORE::ref($dest))); 
    my(%options) = @_[2..$#_]; 
#line 1956 "wmh.pxx"
    
    ###
    # Do the initial setup first
    if ( -l $dest ) {
        unlink $dest;
    } elsif ( -e $dest ) {
        &croak("ERROR: Attempt to make symbolic link but new file '$dest' already exists.");
    }

    ####
    # The most general way to do symlinks is to switch into the
    # directory containing $dest (so we need the dir component of
    # $dest) and symlink from there.  Otherwise, if $src or $dest
    # are not absolute, things are likely to get messed up.

    # Remove trailing '/' from $dest
    $dest =~ s%/$%%;

    # If $src is a file and $dest is a directory, then it means
    # make a file within $dest with the same name as the file
    # portion of $src.  Although the symlink command does this
    # for us automatically, my current algorithm for dealing
    # with all situations requires that we do it explicitly first.
    if ( ( -f $src ) && ( -d $dest ) ) {
        $src =~ m%[^/]+$%;
        my $srcfile = $1;
        $dest .= "/$srcfile";
    }

    # Remove the file portion of $dest (including an optional '/'
    # if the file is actually a directory with trailing '/') to
    # obtain $destdir
    (my $destdir = $dest) =~ s%([^/]+)\s*$%%; $destdir ||= '.';
    my $destfile = $1;
    # In the algorithm below, remember that we will be doing a chdir
    # into $destdir before executing the 'ln -s $somesrc $destfile'
    # command.

    #print("SRC : $src\nDEST: $dest\nDDIR: $destdir\nDF  : $destfile\n");

    if ( $options{relative} ) {
        # We want to change $src to be relative to $dest
        $src = &relativeDirectory($dest,$src);
        #print("SRC : $src\n");
    }

    if ( system("cd $destdir; ln -s $src $destfile") != 0 ) {
        if ( $options{verify} ) {
            die(sprintf("ERORR: %s : $!", ($options{error}||'In Wmh::symlink')));
        }
        return 0;
    } else {
        return 1;
    }
}

#line 2012 "wmh.pxx"
sub write {
    my(%options) = @_[0..$#_]; 
#line 2020 "wmh.pxx"

    my $fp;
    my $file = $options{file};
    if ( $file ) {
        $fp = IO::File->new(">$file");
        if ( ! $fp ) {
            my $error = $options{error} || "Failed to open '$file' for writing.";
            if ( options{die} ) {
                die($error);
            } else {
                printf(STDERR "$error\n");
            }
        }
    } else {
        $fp = $options{fp} || \*STDOUT;
    }

    if ( $options{text} ) {
        print($fp $options{text});
    } elsif ( $options{textref} ) {
        my $ref = $options{textref};
        print($fp $$ref);
    }

}
# User Code
###

###
# Post text

# Post text
###

########################## END OF MODULE ###########################
1;

__DATA__


__END__

=head1 PUBLIC CLASS INTERFACE

=over 4



=head1 PUBLIC INSTANCE INTERFACE

=over 4

=head2 canonicalPhoneNumbers(str $phonelist, str $default_area = '')
Parse a string into a list of phone numbers and canonicalize
each of them (and the list format itself).

=cut

=head2 dataByName(str $name)
Returns a data-structure of the appropriate name ('hash', 'array', etc.)

=cut

=head2 dataFromTextArray(any $data, Array<> $text, any $receiver)
Adds items to datastructure $data based on text in array of lines $text.

For example, if $data is a hash, and $text is the array of the following lines:
|    admins               = HASH
|        bwan@csd.uwo.ca      = 1
|        jchen8@csd.uwo.ca    = 1
|        wade@csd.uwo.ca      = 5
|        wade@gaul.csd.uwo.ca = 5
|        zzwang@csd.uwo.ca    = 1
|    cgifile              = CS307.cgi
|    cgipath              = http://deepwater.csd.uwo.ca:8080/~wade/cgi/CS307.cgi
|    process              = FUNCTION processLoginForm

Then $data will be augmented to have keys 'admins', 'cgifile' and 'cgipath'.

Certain special escapes are possible in the text:
FUNCTION func-name    --> replace func-name with a \& reference to the func-name, which
is assumed to exist in the package in which this method was
evaluated.

=cut

=head2 initState()
Automated method for initializing Perl class state based on field definitions.

=cut

=head2 localFunction(str $func, any $cgiclass)
Given the name of a function and either an actual class object
or the name of the class, return the CODE for $func if it
exists locally.  If it doesn't exist locally, look in parents
NOTE: We currently only look in primary parent!


=cut

=head2 readOneChar()
This method returns a  single character without requiring
a newline to be entered.

NOTE: I'm sure there is a MUCH more efficient way of doing this -
I assume I can change the properties of STDIN with a Fcntl call
to be raw, no?

=cut

=head2 realPath(str $path)
Given an arbitrary path, this routine returns the real path,
by following symbolic links at every possible place.

=cut

=head2 relativeDirectory(str $src, str $dest)
This method returns $dest relative to $src (i.e. with
the appropriate ../.. sequence).  If invoked in list
context, also returns the directory portion of $src
(necessary for symbolic linking to work properly).

=cut

=head2 strToStruct(str $str, ... %options)
Convert a string into a perl ARRAY or HASH.
%options are:
data  --> the structure into which $str should be parsed.
this is only meaningful if $str doesn't start with
a structure-defining word like HASH or ARRAY.
receiver --> for use with FUNCTION specifications, in order to
find an actual function based on its name (starting
in receiver and looking up its inheritance hierarchy
as needed).

=cut

=head2 symlink(str $src, str $dest, ... %options)
Symbolically link real file $str to new file $dest.
If $dest exists and is a symlink, it is removed.
If $dest exists and is not a symlink, an error is generated.
If $src does not exist in $verify is true (default), an error is generated.
Returns 0 on failure, 1 on success

This function use to be less sophisticated, and didn't take
an %options hash, but instead took a single argument whose
semantics are equivalent to $options{verify}.  If you get
errors complaining of odd number of hash elements, it is
because the function is being invoked with 'symlink($src,$dest,1)'
instead of 'symlink($src,$dest,verify=>1)'

%options are:
relative  --> boolean true means try to convert $dest to a path
relative to $dest.
verify    --> boolean true means die on error (instead of just
returning 0)

=cut

=head2 write(... %options)
Options are:
file      --> name of file to write to (errors may occur in opening)
fp        --> IO::Handle instance to write to
text      --> string to print
textref   --> reference to string to print (ignored if text specified)
error     --> string to print out (STDERR) if errors occur
die       --> boolean true means exit on errors

=cut

=head1 AUTHOR

Wade Holst

=head1 SEE ALSO

perl(1).

=cut         